from __future__ import annotations

import pytest


@pytest.fixture()
def primary_ip_response():
    return {
        "primary_ip": {
            "assignee_id": 17,
            "assignee_type": "server",
            "auto_delete": True,
            "blocked": False,
            "created": "2016-01-30T23:55:00+00:00",
            "datacenter": {
                "description": "Falkenstein DC Park 8",
                "id": 42,
                "location": {
                    "city": "Falkenstein",
                    "country": "DE",
                    "description": "Falkenstein DC Park 1",
                    "id": 1,
                    "latitude": 50.47612,
                    "longitude": 12.370071,
                    "name": "fsn1",
                    "network_zone": "eu-central",
                },
                "name": "fsn1-dc8",
                "server_types": {
                    "available": [1, 2, 3],
                    "available_for_migration": [1, 2, 3],
                    "supported": [1, 2, 3],
                },
            },
            "dns_ptr": [{"dns_ptr": "server.example.com", "ip": "131.232.99.1"}],
            "id": 42,
            "ip": "131.232.99.1",
            "labels": {},
            "name": "my-resource",
            "protection": {"delete": False},
            "type": "ipv4",
        }
    }


@pytest.fixture()
def one_primary_ips_response():
    return {
        "meta": {
            "pagination": {
                "last_page": 4,
                "next_page": 4,
                "page": 3,
                "per_page": 25,
                "previous_page": 2,
                "total_entries": 100,
            }
        },
        "primary_ips": [
            {
                "assignee_id": 17,
                "assignee_type": "server",
                "auto_delete": True,
                "blocked": False,
                "created": "2016-01-30T23:55:00+00:00",
                "datacenter": {
                    "description": "Falkenstein DC Park 8",
                    "id": 42,
                    "location": {
                        "city": "Falkenstein",
                        "country": "DE",
                        "description": "Falkenstein DC Park 1",
                        "id": 1,
                        "latitude": 50.47612,
                        "longitude": 12.370071,
                        "name": "fsn1",
                        "network_zone": "eu-central",
                    },
                    "name": "fsn1-dc8",
                    "server_types": {
                        "available": [1, 2, 3],
                        "available_for_migration": [1, 2, 3],
                        "supported": [1, 2, 3],
                    },
                },
                "dns_ptr": [{"dns_ptr": "server.example.com", "ip": "131.232.99.1"}],
                "id": 42,
                "ip": "131.232.99.1",
                "labels": {},
                "name": "my-resource",
                "protection": {"delete": False},
                "type": "ipv4",
            }
        ],
    }


@pytest.fixture()
def all_primary_ips_response():
    return {
        "meta": {
            "pagination": {
                "last_page": 1,
                "next_page": None,
                "page": 1,
                "per_page": 25,
                "previous_page": None,
                "total_entries": 1,
            }
        },
        "primary_ips": [
            {
                "assignee_id": 17,
                "assignee_type": "server",
                "auto_delete": True,
                "blocked": False,
                "created": "2016-01-30T23:55:00+00:00",
                "datacenter": {
                    "description": "Falkenstein DC Park 8",
                    "id": 42,
                    "location": {
                        "city": "Falkenstein",
                        "country": "DE",
                        "description": "Falkenstein DC Park 1",
                        "id": 1,
                        "latitude": 50.47612,
                        "longitude": 12.370071,
                        "name": "fsn1",
                        "network_zone": "eu-central",
                    },
                    "name": "fsn1-dc8",
                    "server_types": {
                        "available": [1, 2, 3],
                        "available_for_migration": [1, 2, 3],
                        "supported": [1, 2, 3],
                    },
                },
                "dns_ptr": [{"dns_ptr": "server.example.com", "ip": "131.232.99.1"}],
                "id": 42,
                "ip": "131.232.99.1",
                "labels": {},
                "name": "my-resource",
                "protection": {"delete": False},
                "type": "ipv4",
            }
        ],
    }


@pytest.fixture()
def primary_ip_create_response():
    return {
        "action": {
            "command": "create_primary_ip",
            "error": {"code": "action_failed", "message": "Action failed"},
            "finished": None,
            "id": 13,
            "progress": 0,
            "resources": [{"id": 17, "type": "server"}],
            "started": "2016-01-30T23:50:00+00:00",
            "status": "running",
        },
        "primary_ip": {
            "assignee_id": 17,
            "assignee_type": "server",
            "auto_delete": True,
            "blocked": False,
            "created": "2016-01-30T23:50:00+00:00",
            "datacenter": {
                "description": "Falkenstein DC Park 8",
                "id": 42,
                "location": {
                    "city": "Falkenstein",
                    "country": "DE",
                    "description": "Falkenstein DC Park 1",
                    "id": 1,
                    "latitude": 50.47612,
                    "longitude": 12.370071,
                    "name": "fsn1",
                    "network_zone": "eu-central",
                    "server_types": {
                        "available": [1, 2, 3],
                        "available_for_migration": [1, 2, 3],
                        "supported": [1, 2, 3],
                    },
                },
                "name": "fsn1-dc8",
            },
            "dns_ptr": [{"dns_ptr": "server.example.com", "ip": "2001:db8::1"}],
            "id": 42,
            "ip": "131.232.99.1",
            "labels": {"labelkey": "value"},
            "name": "my-ip",
            "protection": {"delete": False},
            "type": "ipv4",
        },
    }


@pytest.fixture()
def response_update_primary_ip():
    return {
        "primary_ip": {
            "assignee_id": 17,
            "assignee_type": "server",
            "auto_delete": True,
            "blocked": False,
            "created": "2016-01-30T23:55:00+00:00",
            "datacenter": {
                "description": "Falkenstein DC Park 8",
                "id": 42,
                "location": {
                    "city": "Falkenstein",
                    "country": "DE",
                    "description": "Falkenstein DC Park 1",
                    "id": 1,
                    "latitude": 50.47612,
                    "longitude": 12.370071,
                    "name": "fsn1",
                    "network_zone": "eu-central",
                },
                "name": "fsn1-dc8",
                "server_types": {
                    "available": [1, 2, 3],
                    "available_for_migration": [1, 2, 3],
                    "supported": [1, 2, 3],
                },
            },
            "dns_ptr": [{"dns_ptr": "server.example.com", "ip": "131.232.99.1"}],
            "id": 42,
            "ip": "131.232.99.1",
            "labels": {},
            "name": "my-resource",
            "protection": {"delete": False},
            "type": "ipv4",
        }
    }


@pytest.fixture()
def response_get_actions():
    return {
        "actions": [
            {
                "id": 13,
                "command": "assign_primary_ip",
                "status": "success",
                "progress": 100,
                "started": "2016-01-30T23:55:00+00:00",
                "finished": "2016-01-30T23:56:00+00:00",
                "resources": [{"id": 42, "type": "server"}],
                "error": {"code": "action_failed", "message": "Action failed"},
            }
        ]
    }



from __future__ import annotations

import datetime
from datetime import timezone

from hcloud.servers import Server


class TestServer:
    def test_created_is_datetime(self):
        server = Server(id=1, created="2016-01-30T23:50+00:00")
        assert server.created == datetime.datetime(
            2016, 1, 30, 23, 50, tzinfo=timezone.utc
        )



from __future__ import annotations

import pytest


@pytest.fixture()
def datacenter_response():
    return {
        "datacenter": {
            "id": 1,
            "name": "fsn1-dc8",
            "description": "Falkenstein 1 DC 8",
            "location": {
                "id": 1,
                "name": "fsn1",
                "description": "Falkenstein DC Park 1",
                "country": "DE",
                "city": "Falkenstein",
                "latitude": 50.47612,
                "longitude": 12.370071,
            },
            "server_types": {
                "supported": [1, 2, 3],
                "available": [1, 2, 3],
                "available_for_migration": [1, 2, 3],
            },
        }
    }


@pytest.fixture()
def two_datacenters_response():
    return {
        "datacenters": [
            {
                "id": 1,
                "name": "fsn1-dc8",
                "description": "Falkenstein 1 DC 8",
                "location": {
                    "id": 1,
                    "name": "fsn1",
                    "description": "Falkenstein DC Park 1",
                    "country": "DE",
                    "city": "Falkenstein",
                    "latitude": 50.47612,
                    "longitude": 12.370071,
                },
                "server_types": {
                    "supported": [1, 2, 3],
                    "available": [1, 2, 3],
                    "available_for_migration": [1, 2, 3],
                },
            },
            {
                "id": 2,
                "name": "nbg1-dc3",
                "description": "Nuremberg 1 DC 3",
                "location": {
                    "id": 2,
                    "name": "nbg1",
                    "description": "Nuremberg DC Park 1",
                    "country": "DE",
                    "city": "Nuremberg",
                    "latitude": 49.452102,
                    "longitude": 11.076665,
                },
                "server_types": {
                    "supported": [1, 2, 3],
                    "available": [1, 2, 3],
                    "available_for_migration": [1, 2, 3],
                },
            },
        ],
        "recommendation": 1,
    }


@pytest.fixture()
def one_datacenters_response():
    return {
        "datacenters": [
            {
                "id": 1,
                "name": "fsn1-dc8",
                "description": "Falkenstein 1 DC 8",
                "location": {
                    "id": 1,
                    "name": "fsn1",
                    "description": "Falkenstein DC Park 1",
                    "country": "DE",
                    "city": "Falkenstein",
                    "latitude": 50.47612,
                    "longitude": 12.370071,
                },
                "server_types": {
                    "supported": [1, 2, 3],
                    "available": [1, 2, 3],
                    "available_for_migration": [1, 2, 3],
                },
            }
        ],
        "recommendation": 1,
    }



from __future__ import annotations

import warnings

warnings.warn(
    "The 'hcloud.hcloud' module is deprecated, please import from the 'hcloud' module instead (e.g. 'from hcloud import Client').",
    DeprecationWarning,
    stacklevel=2,
)

# pylint: disable=wildcard-import,wrong-import-position,unused-wildcard-import
from ._client import *  # noqa



from __future__ import annotations

from unittest import mock  # noqa: F401

import pytest  # noqa: F401

from hcloud.datacenters import BoundDatacenter, DatacentersClient, DatacenterServerTypes
from hcloud.locations import BoundLocation


class TestBoundDatacenter:
    def test_bound_datacenter_init(self, datacenter_response):
        bound_datacenter = BoundDatacenter(
            client=mock.MagicMock(), data=datacenter_response["datacenter"]
        )

        assert bound_datacenter.id == 1
        assert bound_datacenter.name == "fsn1-dc8"
        assert bound_datacenter.description == "Falkenstein 1 DC 8"
        assert bound_datacenter.complete is True

        assert isinstance(bound_datacenter.location, BoundLocation)
        assert bound_datacenter.location.id == 1
        assert bound_datacenter.location.name == "fsn1"
        assert bound_datacenter.location.complete is True

        assert isinstance(bound_datacenter.server_types, DatacenterServerTypes)
        assert len(bound_datacenter.server_types.supported) == 3
        assert bound_datacenter.server_types.supported[0].id == 1
        assert bound_datacenter.server_types.supported[0].complete is False
        assert bound_datacenter.server_types.supported[1].id == 2
        assert bound_datacenter.server_types.supported[1].complete is False
        assert bound_datacenter.server_types.supported[2].id == 3
        assert bound_datacenter.server_types.supported[2].complete is False

        assert len(bound_datacenter.server_types.available) == 3
        assert bound_datacenter.server_types.available[0].id == 1
        assert bound_datacenter.server_types.available[0].complete is False
        assert bound_datacenter.server_types.available[1].id == 2
        assert bound_datacenter.server_types.available[1].complete is False
        assert bound_datacenter.server_types.available[2].id == 3
        assert bound_datacenter.server_types.available[2].complete is False

        assert len(bound_datacenter.server_types.available_for_migration) == 3
        assert bound_datacenter.server_types.available_for_migration[0].id == 1
        assert (
            bound_datacenter.server_types.available_for_migration[0].complete is False
        )
        assert bound_datacenter.server_types.available_for_migration[1].id == 2
        assert (
            bound_datacenter.server_types.available_for_migration[1].complete is False
        )
        assert bound_datacenter.server_types.available_for_migration[2].id == 3
        assert (
            bound_datacenter.server_types.available_for_migration[2].complete is False
        )


class TestDatacentersClient:
    @pytest.fixture()
    def datacenters_client(self):
        return DatacentersClient(client=mock.MagicMock())

    def test_get_by_id(self, datacenters_client, datacenter_response):
        datacenters_client._client.request.return_value = datacenter_response
        datacenter = datacenters_client.get_by_id(1)
        datacenters_client._client.request.assert_called_with(
            url="/datacenters/1", method="GET"
        )
        assert datacenter._client is datacenters_client
        assert datacenter.id == 1
        assert datacenter.name == "fsn1-dc8"

    @pytest.mark.parametrize(
        "params", [{"name": "fsn1", "page": 1, "per_page": 10}, {"name": ""}, {}]
    )
    def test_get_list(self, datacenters_client, two_datacenters_response, params):
        datacenters_client._client.request.return_value = two_datacenters_response
        result = datacenters_client.get_list(**params)
        datacenters_client._client.request.assert_called_with(
            url="/datacenters", method="GET", params=params
        )

        datacenters = result.datacenters
        assert result.meta is None

        assert len(datacenters) == 2

        datacenter1 = datacenters[0]
        datacenter2 = datacenters[1]

        assert datacenter1._client is datacenters_client
        assert datacenter1.id == 1
        assert datacenter1.name == "fsn1-dc8"
        assert isinstance(datacenter1.location, BoundLocation)

        assert datacenter2._client is datacenters_client
        assert datacenter2.id == 2
        assert datacenter2.name == "nbg1-dc3"
        assert isinstance(datacenter2.location, BoundLocation)

    @pytest.mark.parametrize("params", [{"name": "fsn1"}, {}])
    def test_get_all(self, datacenters_client, two_datacenters_response, params):
        datacenters_client._client.request.return_value = two_datacenters_response
        datacenters = datacenters_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})
        datacenters_client._client.request.assert_called_with(
            url="/datacenters", method="GET", params=params
        )

        assert len(datacenters) == 2

        datacenter1 = datacenters[0]
        datacenter2 = datacenters[1]

        assert datacenter1._client is datacenters_client
        assert datacenter1.id == 1
        assert datacenter1.name == "fsn1-dc8"
        assert isinstance(datacenter1.location, BoundLocation)

        assert datacenter2._client is datacenters_client
        assert datacenter2.id == 2
        assert datacenter2.name == "nbg1-dc3"
        assert isinstance(datacenter2.location, BoundLocation)

    def test_get_by_name(self, datacenters_client, one_datacenters_response):
        datacenters_client._client.request.return_value = one_datacenters_response
        datacenter = datacenters_client.get_by_name("fsn1-dc8")

        params = {"name": "fsn1-dc8"}
        datacenters_client._client.request.assert_called_with(
            url="/datacenters", method="GET", params=params
        )

        assert datacenter._client is datacenters_client
        assert datacenter.id == 1
        assert datacenter.name == "fsn1-dc8"
        assert isinstance(datacenter.location, BoundLocation)



from __future__ import annotations

from datetime import datetime, timezone
from unittest import mock

import pytest

from hcloud.server_types import BoundServerType, ServerTypesClient


class TestBoundServerType:
    @pytest.fixture()
    def bound_server_type(self, hetzner_client):
        return BoundServerType(client=hetzner_client.server_types, data=dict(id=14))

    def test_bound_server_type_init(self, server_type_response):
        bound_server_type = BoundServerType(
            client=mock.MagicMock(), data=server_type_response["server_type"]
        )

        assert bound_server_type.id == 1
        assert bound_server_type.name == "cx11"
        assert bound_server_type.description == "CX11"
        assert bound_server_type.cores == 1
        assert bound_server_type.memory == 1
        assert bound_server_type.disk == 25
        assert bound_server_type.storage_type == "local"
        assert bound_server_type.cpu_type == "shared"
        assert bound_server_type.architecture == "x86"
        assert bound_server_type.deprecated is True
        assert bound_server_type.deprecation is not None
        assert bound_server_type.deprecation.announced == datetime(
            2023, 6, 1, tzinfo=timezone.utc
        )
        assert bound_server_type.deprecation.unavailable_after == datetime(
            2023, 9, 1, tzinfo=timezone.utc
        )
        assert bound_server_type.included_traffic == 21990232555520


class TestServerTypesClient:
    @pytest.fixture()
    def server_types_client(self):
        return ServerTypesClient(client=mock.MagicMock())

    def test_get_by_id(self, server_types_client, server_type_response):
        server_types_client._client.request.return_value = server_type_response
        server_type = server_types_client.get_by_id(1)
        server_types_client._client.request.assert_called_with(
            url="/server_types/1", method="GET"
        )
        assert server_type._client is server_types_client
        assert server_type.id == 1
        assert server_type.name == "cx11"

    @pytest.mark.parametrize(
        "params", [{"name": "cx11", "page": 1, "per_page": 10}, {"name": ""}, {}]
    )
    def test_get_list(self, server_types_client, two_server_types_response, params):
        server_types_client._client.request.return_value = two_server_types_response
        result = server_types_client.get_list(**params)
        server_types_client._client.request.assert_called_with(
            url="/server_types", method="GET", params=params
        )

        server_types = result.server_types
        assert result.meta is None

        assert len(server_types) == 2

        server_types1 = server_types[0]
        server_types2 = server_types[1]

        assert server_types1._client is server_types_client
        assert server_types1.id == 1
        assert server_types1.name == "cx11"

        assert server_types2._client is server_types_client
        assert server_types2.id == 2
        assert server_types2.name == "cx21"

    @pytest.mark.parametrize("params", [{"name": "cx11"}])
    def test_get_all(self, server_types_client, two_server_types_response, params):
        server_types_client._client.request.return_value = two_server_types_response
        server_types = server_types_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})

        server_types_client._client.request.assert_called_with(
            url="/server_types", method="GET", params=params
        )

        assert len(server_types) == 2

        server_types1 = server_types[0]
        server_types2 = server_types[1]

        assert server_types1._client is server_types_client
        assert server_types1.id == 1
        assert server_types1.name == "cx11"

        assert server_types2._client is server_types_client
        assert server_types2.id == 2
        assert server_types2.name == "cx21"

    def test_get_by_name(self, server_types_client, one_server_types_response):
        server_types_client._client.request.return_value = one_server_types_response
        server_type = server_types_client.get_by_name("cx11")

        params = {"name": "cx11"}

        server_types_client._client.request.assert_called_with(
            url="/server_types", method="GET", params=params
        )

        assert server_type._client is server_types_client
        assert server_type.id == 1
        assert server_type.name == "cx11"



from __future__ import annotations

from .client import BoundFirewall, FirewallsClient, FirewallsPageResult  # noqa: F401
from .domain import (  # noqa: F401
    CreateFirewallResponse,
    Firewall,
    FirewallResource,
    FirewallResourceAppliedToResources,
    FirewallResourceLabelSelector,
    FirewallRule,
)



from __future__ import annotations

from ..core import BaseDomain, DomainIdentityMixin


class LoadBalancerType(BaseDomain, DomainIdentityMixin):
    """LoadBalancerType Domain

    :param id: int
           ID of the Load Balancer type
    :param name: str
           Name of the Load Balancer type
    :param description: str
           Description of the Load Balancer type
    :param max_connections: int
           Max amount of connections the Load Balancer can handle
    :param max_services: int
           Max amount of services the Load Balancer can handle
    :param max_targets: int
           Max amount of targets the Load Balancer can handle
    :param max_assigned_certificates: int
           Max amount of certificates the Load Balancer can serve
    :param prices: Dict
           Prices in different locations

    """

    __slots__ = (
        "id",
        "name",
        "description",
        "max_connections",
        "max_services",
        "max_targets",
        "max_assigned_certificates",
        "prices",
    )

    def __init__(
        self,
        id: int | None = None,
        name: str | None = None,
        description: str | None = None,
        max_connections: int | None = None,
        max_services: int | None = None,
        max_targets: int | None = None,
        max_assigned_certificates: int | None = None,
        prices: dict | None = None,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.max_connections = max_connections
        self.max_services = max_services
        self.max_targets = max_targets
        self.max_assigned_certificates = max_assigned_certificates
        self.prices = prices



from __future__ import annotations

import pytest


@pytest.fixture()
def certificate_response():
    return {
        "certificate": {
            "id": 2323,
            "name": "My Certificate",
            "type": "managed",
            "labels": {},
            "certificate": "-----BEGIN CERTIFICATE-----\n...",
            "created": "2019-01-08T12:10:00+00:00",
            "not_valid_before": "2019-01-08T10:00:00+00:00",
            "not_valid_after": "2019-07-08T09:59:59+00:00",
            "domain_names": ["example.com", "webmail.example.com", "www.example.com"],
            "fingerprint": "03:c7:55:9b:2a:d1:04:17:09:f6:d0:7f:18:34:63:d4:3e:5f",
            "status": {
                "issuance": "failed",
                "renewal": "scheduled",
                "error": {"code": "error_code", "message": "error message"},
            },
            "used_by": [{"id": 42, "type": "server"}],
        }
    }


@pytest.fixture()
def create_managed_certificate_response():
    return {
        "certificate": {
            "id": 2323,
            "name": "My Certificate",
            "type": "managed",
            "labels": {},
            "certificate": "-----BEGIN CERTIFICATE-----\n...",
            "created": "2019-01-08T12:10:00+00:00",
            "not_valid_before": "2019-01-08T10:00:00+00:00",
            "not_valid_after": "2019-07-08T09:59:59+00:00",
            "domain_names": ["example.com", "webmail.example.com", "www.example.com"],
            "fingerprint": "03:c7:55:9b:2a:d1:04:17:09:f6:d0:7f:18:34:63:d4:3e:5f",
            "status": {"issuance": "pending", "renewal": "scheduled", "error": None},
            "used_by": [{"id": 42, "type": "load_balancer"}],
        },
        "action": {
            "id": 14,
            "command": "issue_certificate",
            "status": "success",
            "progress": 100,
            "started": "2021-01-30T23:55:00+00:00",
            "finished": "2021-01-30T23:57:00+00:00",
            "resources": [{"id": 896, "type": "certificate"}],
            "error": {"code": "action_failed", "message": "Action failed"},
        },
    }


@pytest.fixture()
def two_certificates_response():
    return {
        "certificates": [
            {
                "id": 2323,
                "name": "My Certificate",
                "labels": {},
                "type": "uploaded",
                "certificate": "-----BEGIN CERTIFICATE-----\n...",
                "created": "2019-01-08T12:10:00+00:00",
                "not_valid_before": "2019-01-08T10:00:00+00:00",
                "not_valid_after": "2019-07-08T09:59:59+00:00",
                "domain_names": [
                    "example.com",
                    "webmail.example.com",
                    "www.example.com",
                ],
                "fingerprint": "03:c7:55:9b:2a:d1:04:17:09:f6:d0:7f:18:34:63:d4:3e:5f",
                "status": None,
                "used_by": [{"id": 42, "type": "load_balancer"}],
            },
            {
                "id": 2324,
                "name": "My website cert",
                "labels": {},
                "type": "uploaded",
                "certificate": "-----BEGIN CERTIFICATE-----\n...",
                "created": "2019-01-08T12:10:00+00:00",
                "not_valid_before": "2019-01-08T10:00:00+00:00",
                "not_valid_after": "2019-07-08T09:59:59+00:00",
                "domain_names": [
                    "example.com",
                    "webmail.example.com",
                    "www.example.com",
                ],
                "fingerprint": "03:c7:55:9b:2a:d1:04:17:09:f6:d0:7f:18:34:63:d4:3e:5f",
                "status": None,
                "used_by": [{"id": 42, "type": "load_balancer"}],
            },
        ]
    }


@pytest.fixture()
def one_certificates_response():
    return {
        "certificates": [
            {
                "id": 2323,
                "name": "My Certificate",
                "labels": {},
                "type": "uploaded",
                "certificate": "-----BEGIN CERTIFICATE-----\n...",
                "created": "2019-01-08T12:10:00+00:00",
                "not_valid_before": "2019-01-08T10:00:00+00:00",
                "not_valid_after": "2019-07-08T09:59:59+00:00",
                "domain_names": [
                    "example.com",
                    "webmail.example.com",
                    "www.example.com",
                ],
                "fingerprint": "03:c7:55:9b:2a:d1:04:17:09:f6:d0:7f:18:34:63:d4:3e:5f",
                "status": None,
                "used_by": [{"id": 42, "type": "load_balancer"}],
            }
        ]
    }


@pytest.fixture()
def response_update_certificate():
    return {
        "certificate": {
            "id": 2323,
            "name": "New name",
            "labels": {},
            "type": "uploaded",
            "certificate": "-----BEGIN CERTIFICATE-----\n...",
            "created": "2019-01-08T12:10:00+00:00",
            "not_valid_before": "2019-01-08T10:00:00+00:00",
            "not_valid_after": "2019-07-08T09:59:59+00:00",
            "domain_names": ["example.com", "webmail.example.com", "www.example.com"],
            "fingerprint": "03:c7:55:9b:2a:d1:04:17:09:f6:d0:7f:18:34:63:d4:3e:5f",
            "status": None,
            "used_by": [{"id": 42, "type": "load_balancer"}],
        }
    }


@pytest.fixture()
def response_get_actions():
    return {
        "actions": [
            {
                "id": 13,
                "command": "change_protection",
                "status": "success",
                "progress": 100,
                "started": "2016-01-30T23:55:00+00:00",
                "finished": "2016-01-30T23:56:00+00:00",
                "resources": [{"id": 14, "type": "certificate"}],
                "error": {"code": "action_failed", "message": "Action failed"},
            }
        ]
    }


@pytest.fixture()
def response_retry_issuance_action():
    return {
        "action": {
            "id": 14,
            "command": "issue_certificate",
            "status": "running",
            "progress": 0,
            "started": "2016-01-30T23:50+00:00",
            "finished": None,
            "resources": [{"id": 42, "type": "certificate"}],
            "error": {"code": "action_failed", "message": "Action failed"},
        }
    }



from __future__ import annotations

from datetime import datetime
from typing import Dict, List, Literal, Tuple

from dateutil.parser import isoparse

from ..core import BaseDomain

TimeSeries = Dict[str, Dict[Literal["values"], List[Tuple[float, str]]]]


class Metrics(BaseDomain):
    """Metrics Domain

    :param start: Start of period of metrics reported.
    :param end: End of period of metrics reported.
    :param step: Resolution of results in seconds.
    :param time_series: Dict with time series data, using the name of the time series as
        key. The metrics timestamps and values are stored in a list of tuples
        ``[(timestamp, value), ...]``.
    """

    start: datetime
    end: datetime
    step: float
    time_series: TimeSeries

    __slots__ = (
        "start",
        "end",
        "step",
        "time_series",
    )

    def __init__(
        self,
        start: str,
        end: str,
        step: float,
        time_series: TimeSeries,
    ):
        self.start = isoparse(start)
        self.end = isoparse(end)
        self.step = step
        self.time_series = time_series



from __future__ import annotations

from typing import TYPE_CHECKING, Any, NamedTuple

from ..core import BoundModelBase, ClientEntityBase, Meta
from .domain import SSHKey

if TYPE_CHECKING:
    from .._client import Client


class BoundSSHKey(BoundModelBase, SSHKey):
    _client: SSHKeysClient

    model = SSHKey

    def update(
        self,
        name: str | None = None,
        labels: dict[str, str] | None = None,
    ) -> BoundSSHKey:
        """Updates an SSH key. You can update an SSH key name and an SSH key labels.

        :param description: str (optional)
               New Description to set
        :param labels: Dict[str, str] (optional)
               User-defined labels (key-value pairs)
        :return: :class:`BoundSSHKey <hcloud.ssh_keys.client.BoundSSHKey>`
        """
        return self._client.update(self, name, labels)

    def delete(self) -> bool:
        """Deletes an SSH key. It cannot be used anymore.
        :return: boolean
        """
        return self._client.delete(self)


class SSHKeysPageResult(NamedTuple):
    ssh_keys: list[BoundSSHKey]
    meta: Meta | None


class SSHKeysClient(ClientEntityBase):
    _client: Client

    def get_by_id(self, id: int) -> BoundSSHKey:
        """Get a specific SSH Key by its ID

        :param id: int
        :return: :class:`BoundSSHKey <hcloud.ssh_keys.client.BoundSSHKey>`
        """
        response = self._client.request(url=f"/ssh_keys/{id}", method="GET")
        return BoundSSHKey(self, response["ssh_key"])

    def get_list(
        self,
        name: str | None = None,
        fingerprint: str | None = None,
        label_selector: str | None = None,
        page: int | None = None,
        per_page: int | None = None,
    ) -> SSHKeysPageResult:
        """Get a list of SSH keys from the account

        :param name: str (optional)
               Can be used to filter SSH keys by their name. The response will only contain the SSH key matching the specified name.
        :param fingerprint: str (optional)
               Can be used to filter SSH keys by their fingerprint. The response will only contain the SSH key matching the specified fingerprint.
        :param label_selector: str (optional)
               Can be used to filter SSH keys by labels. The response will only contain SSH keys matching the label selector.
        :param page: int (optional)
               Specifies the page to fetch
        :param per_page: int (optional)
               Specifies how many results are returned by page
        :return:  (List[:class:`BoundSSHKey <hcloud.ssh_keys.client.BoundSSHKey>`], :class:`Meta <hcloud.core.domain.Meta>`)
        """
        params: dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        if fingerprint is not None:
            params["fingerprint"] = fingerprint
        if label_selector is not None:
            params["label_selector"] = label_selector
        if page is not None:
            params["page"] = page
        if per_page is not None:
            params["per_page"] = per_page

        response = self._client.request(url="/ssh_keys", method="GET", params=params)

        ssh_keys = [
            BoundSSHKey(self, server_data) for server_data in response["ssh_keys"]
        ]
        return SSHKeysPageResult(ssh_keys, Meta.parse_meta(response))

    def get_all(
        self,
        name: str | None = None,
        fingerprint: str | None = None,
        label_selector: str | None = None,
    ) -> list[BoundSSHKey]:
        """Get all SSH keys from the account

        :param name: str (optional)
               Can be used to filter SSH keys by their name. The response will only contain the SSH key matching the specified name.
        :param fingerprint: str (optional)
               Can be used to filter SSH keys by their fingerprint. The response will only contain the SSH key matching the specified fingerprint.
        :param label_selector: str (optional)
               Can be used to filter SSH keys by labels. The response will only contain SSH keys matching the label selector.
        :return:  List[:class:`BoundSSHKey <hcloud.ssh_keys.client.BoundSSHKey>`]
        """
        return self._iter_pages(
            self.get_list,
            name=name,
            fingerprint=fingerprint,
            label_selector=label_selector,
        )

    def get_by_name(self, name: str) -> BoundSSHKey | None:
        """Get ssh key by name

        :param name: str
               Used to get ssh key by name.
        :return: :class:`BoundSSHKey <hcloud.ssh_keys.client.BoundSSHKey>`
        """
        return self._get_first_by(name=name)

    def get_by_fingerprint(self, fingerprint: str) -> BoundSSHKey | None:
        """Get ssh key by fingerprint

        :param fingerprint: str
                Used to get ssh key by fingerprint.
        :return: :class:`BoundSSHKey <hcloud.ssh_keys.client.BoundSSHKey>`
        """
        return self._get_first_by(fingerprint=fingerprint)

    def create(
        self,
        name: str,
        public_key: str,
        labels: dict[str, str] | None = None,
    ) -> BoundSSHKey:
        """Creates a new SSH key with the given name and public_key.

        :param name: str
        :param public_key: str
               Public Key of the SSH Key you want create
        :param labels: Dict[str, str] (optional)
               User-defined labels (key-value pairs)
        :return: :class:`BoundSSHKey <hcloud.ssh_keys.client.BoundSSHKey>`
        """
        data: dict[str, Any] = {"name": name, "public_key": public_key}
        if labels is not None:
            data["labels"] = labels
        response = self._client.request(url="/ssh_keys", method="POST", json=data)
        return BoundSSHKey(self, response["ssh_key"])

    def update(
        self,
        ssh_key: SSHKey | BoundSSHKey,
        name: str | None = None,
        labels: dict[str, str] | None = None,
    ) -> BoundSSHKey:
        """Updates an SSH key. You can update an SSH key name and an SSH key labels.

        :param ssh_key: :class:`BoundSSHKey <hcloud.ssh_keys.client.BoundSSHKey>` or  :class:`SSHKey <hcloud.ssh_keys.domain.SSHKey>`
        :param name: str (optional)
               New Description to set
        :param labels: Dict[str, str] (optional)
               User-defined labels (key-value pairs)
        :return: :class:`BoundSSHKey <hcloud.ssh_keys.client.BoundSSHKey>`
        """
        data: dict[str, Any] = {}
        if name is not None:
            data["name"] = name
        if labels is not None:
            data["labels"] = labels
        response = self._client.request(
            url=f"/ssh_keys/{ssh_key.id}",
            method="PUT",
            json=data,
        )
        return BoundSSHKey(self, response["ssh_key"])

    def delete(self, ssh_key: SSHKey | BoundSSHKey) -> bool:
        """Deletes an SSH key. It cannot be used anymore.

        :param ssh_key: :class:`BoundSSHKey <hcloud.ssh_keys.client.BoundSSHKey>` or  :class:`SSHKey <hcloud.ssh_keys.domain.SSHKey>`
        :return: True
        """
        self._client.request(url=f"/ssh_keys/{ssh_key.id}", method="DELETE")
        # Return always true, because the API does not return an action for it. When an error occurs a HcloudAPIException will be raised
        return True



from __future__ import annotations

import datetime
from datetime import timezone

from hcloud.load_balancers import LoadBalancer


class TestLoadBalancers:
    def test_created_is_datetime(self):
        lb = LoadBalancer(id=1, created="2016-01-30T23:50+00:00")
        assert lb.created == datetime.datetime(2016, 1, 30, 23, 50, tzinfo=timezone.utc)



from __future__ import annotations

import pytest


@pytest.fixture()
def floating_ip_response():
    return {
        "floating_ip": {
            "id": 4711,
            "description": "Web Frontend",
            "name": "Web Frontend",
            "created": "2016-01-30T23:50+00:00",
            "ip": "131.232.99.1",
            "type": "ipv4",
            "server": 42,
            "dns_ptr": [{"ip": "2001:db8::1", "dns_ptr": "server.example.com"}],
            "home_location": {
                "id": 1,
                "name": "fsn1",
                "description": "Falkenstein DC Park 1",
                "country": "DE",
                "city": "Falkenstein",
                "latitude": 50.47612,
                "longitude": 12.370071,
            },
            "blocked": False,
            "protection": {"delete": False},
            "labels": {},
        }
    }


@pytest.fixture()
def one_floating_ips_response():
    return {
        "floating_ips": [
            {
                "id": 4711,
                "description": "Web Frontend",
                "name": "Web Frontend",
                "created": "2016-01-30T23:50+00:00",
                "ip": "131.232.99.1",
                "type": "ipv4",
                "server": 42,
                "dns_ptr": [{"ip": "2001:db8::1", "dns_ptr": "server.example.com"}],
                "home_location": {
                    "id": 1,
                    "name": "fsn1",
                    "description": "Falkenstein DC Park 1",
                    "country": "DE",
                    "city": "Falkenstein",
                    "latitude": 50.47612,
                    "longitude": 12.370071,
                },
                "blocked": False,
                "protection": {"delete": False},
                "labels": {},
            }
        ]
    }


@pytest.fixture()
def two_floating_ips_response():
    return {
        "floating_ips": [
            {
                "id": 4711,
                "description": "Web Frontend",
                "name": "Web Frontend",
                "created": "2016-01-30T23:50+00:00",
                "ip": "131.232.99.1",
                "type": "ipv4",
                "server": 42,
                "dns_ptr": [{"ip": "2001:db8::1", "dns_ptr": "server.example.com"}],
                "home_location": {
                    "id": 1,
                    "name": "fsn1",
                    "description": "Falkenstein DC Park 1",
                    "country": "DE",
                    "city": "Falkenstein",
                    "latitude": 50.47612,
                    "longitude": 12.370071,
                },
                "blocked": False,
                "protection": {"delete": False},
                "labels": {},
            },
            {
                "id": 4712,
                "description": "Web Backend",
                "name": "Web Backend",
                "created": "2016-01-30T23:50+00:00",
                "ip": "131.232.99.2",
                "type": "ipv4",
                "server": 42,
                "dns_ptr": [{"ip": "2001:db8::1", "dns_ptr": "server.example.com"}],
                "home_location": {
                    "id": 1,
                    "name": "fsn1",
                    "description": "Falkenstein DC Park 1",
                    "country": "DE",
                    "city": "Falkenstein",
                    "latitude": 50.47612,
                    "longitude": 12.370071,
                },
                "blocked": False,
                "protection": {"delete": False},
                "labels": {},
            },
        ]
    }


@pytest.fixture()
def floating_ip_create_response():
    return {
        "floating_ip": {
            "id": 4711,
            "description": "Web Frontend",
            "name": "Web Frontend",
            "created": "2016-01-30T23:50+00:00",
            "ip": "131.232.99.1",
            "type": "ipv4",
            "server": 42,
            "dns_ptr": [{"ip": "2001:db8::1", "dns_ptr": "server.example.com"}],
            "home_location": {
                "id": 1,
                "name": "fsn1",
                "description": "Falkenstein DC Park 1",
                "country": "DE",
                "city": "Falkenstein",
                "latitude": 50.47612,
                "longitude": 12.370071,
            },
            "blocked": False,
            "protection": {"delete": False},
            "labels": {},
        },
        "action": {
            "id": 13,
            "command": "assign_floating_ip",
            "status": "running",
            "progress": 0,
            "started": "2016-01-30T23:50+00:00",
            "finished": None,
            "resources": [{"id": 42, "type": "server"}],
            "error": {"code": "action_failed", "message": "Action failed"},
        },
    }


@pytest.fixture()
def response_update_floating_ip():
    return {
        "floating_ip": {
            "id": 4711,
            "description": "New description",
            "name": "New name",
            "created": "2016-01-30T23:50+00:00",
            "ip": "131.232.99.1",
            "type": "ipv4",
            "server": 42,
            "dns_ptr": [{"ip": "2001:db8::1", "dns_ptr": "server.example.com"}],
            "home_location": {
                "id": 1,
                "name": "fsn1",
                "description": "Falkenstein DC Park 1",
                "country": "DE",
                "city": "Falkenstein",
                "latitude": 50.47612,
                "longitude": 12.370071,
            },
            "blocked": False,
            "protection": {"delete": False},
            "labels": {},
        }
    }


@pytest.fixture()
def response_get_actions():
    return {
        "actions": [
            {
                "id": 13,
                "command": "assign_floating_ip",
                "status": "success",
                "progress": 100,
                "started": "2016-01-30T23:55:00+00:00",
                "finished": "2016-01-30T23:56:00+00:00",
                "resources": [{"id": 42, "type": "server"}],
                "error": {"code": "action_failed", "message": "Action failed"},
            }
        ]
    }



from __future__ import annotations

from unittest import mock

import pytest

from hcloud.actions import BoundAction
from hcloud.load_balancer_types import LoadBalancerType
from hcloud.load_balancers import (
    BoundLoadBalancer,
    LoadBalancer,
    LoadBalancerAlgorithm,
    LoadBalancerHealthCheck,
    LoadBalancersClient,
    LoadBalancerService,
    LoadBalancerTarget,
    LoadBalancerTargetIP,
    LoadBalancerTargetLabelSelector,
)
from hcloud.locations import Location
from hcloud.networks import Network
from hcloud.servers import Server


class TestBoundLoadBalancer:
    @pytest.fixture()
    def bound_load_balancer(self, hetzner_client):
        return BoundLoadBalancer(client=hetzner_client.load_balancers, data=dict(id=14))

    def test_bound_load_balancer_init(self, response_load_balancer):
        bound_load_balancer = BoundLoadBalancer(
            client=mock.MagicMock(), data=response_load_balancer["load_balancer"]
        )

        assert bound_load_balancer.id == 4711
        assert bound_load_balancer.name == "Web Frontend"

    @pytest.mark.parametrize("params", [{"page": 1, "per_page": 10}, {}])
    def test_get_actions_list(
        self, hetzner_client, bound_load_balancer, response_get_actions, params
    ):
        hetzner_client.request.return_value = response_get_actions
        result = bound_load_balancer.get_actions_list(**params)
        hetzner_client.request.assert_called_with(
            url="/load_balancers/14/actions", method="GET", params=params
        )

        actions = result.actions
        assert result.meta is None

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == hetzner_client.actions
        assert actions[0].id == 13
        assert actions[0].command == "change_protection"

    @pytest.mark.parametrize("params", [{}])
    def test_get_actions(
        self, hetzner_client, bound_load_balancer, response_get_actions, params
    ):
        hetzner_client.request.return_value = response_get_actions
        actions = bound_load_balancer.get_actions(**params)

        params.update({"page": 1, "per_page": 50})

        hetzner_client.request.assert_called_with(
            url="/load_balancers/14/actions", method="GET", params=params
        )

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == hetzner_client.actions
        assert actions[0].id == 13
        assert actions[0].command == "change_protection"

    def test_update(
        self, hetzner_client, bound_load_balancer, response_update_load_balancer
    ):
        hetzner_client.request.return_value = response_update_load_balancer
        load_balancer = bound_load_balancer.update(name="new-name", labels={})
        hetzner_client.request.assert_called_with(
            url="/load_balancers/14",
            method="PUT",
            json={"name": "new-name", "labels": {}},
        )

        assert load_balancer.id == 4711
        assert load_balancer.name == "new-name"

    def test_delete(self, hetzner_client, generic_action, bound_load_balancer):
        hetzner_client.request.return_value = generic_action
        delete_success = bound_load_balancer.delete()
        hetzner_client.request.assert_called_with(
            url="/load_balancers/14", method="DELETE"
        )

        assert delete_success is True

    def test_get_metrics(
        self,
        hetzner_client,
        response_get_metrics,
        bound_load_balancer: BoundLoadBalancer,
    ):
        hetzner_client.request.return_value = response_get_metrics
        response = bound_load_balancer.get_metrics(
            type=["requests_per_second"],
            start="2023-12-14T16:55:32+01:00",
            end="2023-12-14T16:55:32+01:00",
        )
        hetzner_client.request.assert_called_with(
            url="/load_balancers/14/metrics",
            method="GET",
            params={
                "type": "requests_per_second",
                "start": "2023-12-14T16:55:32+01:00",
                "end": "2023-12-14T16:55:32+01:00",
            },
        )
        assert "requests_per_second" in response.metrics.time_series
        assert len(response.metrics.time_series["requests_per_second"]["values"]) == 3

    def test_add_service(
        self, hetzner_client, response_add_service, bound_load_balancer
    ):
        hetzner_client.request.return_value = response_add_service
        service = LoadBalancerService(listen_port=80, protocol="http")
        action = bound_load_balancer.add_service(service)
        hetzner_client.request.assert_called_with(
            json={"protocol": "http", "listen_port": 80},
            url="/load_balancers/14/actions/add_service",
            method="POST",
        )

        assert action.id == 13
        assert action.progress == 100
        assert action.command == "add_service"

    def test_delete_service(
        self, hetzner_client, response_delete_service, bound_load_balancer
    ):
        hetzner_client.request.return_value = response_delete_service
        service = LoadBalancerService(listen_port=12)
        action = bound_load_balancer.delete_service(service)
        hetzner_client.request.assert_called_with(
            json={"listen_port": 12},
            url="/load_balancers/14/actions/delete_service",
            method="POST",
        )

        assert action.id == 13
        assert action.progress == 100
        assert action.command == "delete_service"

    @pytest.mark.parametrize(
        "target,params",
        [
            (
                LoadBalancerTarget(
                    type="server", server=Server(id=1), use_private_ip=True
                ),
                {"server": {"id": 1}, "use_private_ip": True},
            ),
            (
                LoadBalancerTarget(type="ip", ip=LoadBalancerTargetIP(ip="127.0.0.1")),
                {"ip": {"ip": "127.0.0.1"}},
            ),
            (
                LoadBalancerTarget(
                    type="label_selector",
                    label_selector=LoadBalancerTargetLabelSelector(selector="abc=def"),
                ),
                {"label_selector": {"selector": "abc=def"}},
            ),
        ],
    )
    def test_add_target(
        self, hetzner_client, response_add_target, bound_load_balancer, target, params
    ):
        hetzner_client.request.return_value = response_add_target
        action = bound_load_balancer.add_target(target)
        params.update({"type": target.type})
        hetzner_client.request.assert_called_with(
            url="/load_balancers/14/actions/add_target", method="POST", json=params
        )

        assert action.id == 13
        assert action.progress == 100
        assert action.command == "add_target"

    @pytest.mark.parametrize(
        "target,params",
        [
            (
                LoadBalancerTarget(
                    type="server", server=Server(id=1), use_private_ip=True
                ),
                {"server": {"id": 1}},
            ),
            (
                LoadBalancerTarget(type="ip", ip=LoadBalancerTargetIP(ip="127.0.0.1")),
                {"ip": {"ip": "127.0.0.1"}},
            ),
            (
                LoadBalancerTarget(
                    type="label_selector",
                    label_selector=LoadBalancerTargetLabelSelector(selector="abc=def"),
                ),
                {"label_selector": {"selector": "abc=def"}},
            ),
        ],
    )
    def test_remove_target(
        self,
        hetzner_client,
        response_remove_target,
        bound_load_balancer,
        target,
        params,
    ):
        hetzner_client.request.return_value = response_remove_target
        action = bound_load_balancer.remove_target(target)
        params.update({"type": target.type})
        hetzner_client.request.assert_called_with(
            url="/load_balancers/14/actions/remove_target", method="POST", json=params
        )

        assert action.id == 13
        assert action.progress == 100
        assert action.command == "remove_target"

    def test_update_service(
        self, hetzner_client, response_update_service, bound_load_balancer
    ):
        hetzner_client.request.return_value = response_update_service
        new_health_check = LoadBalancerHealthCheck(
            protocol="http", port=13, interval=1, timeout=1, retries=1
        )
        service = LoadBalancerService(listen_port=12, health_check=new_health_check)

        action = bound_load_balancer.update_service(service)
        hetzner_client.request.assert_called_with(
            json={
                "listen_port": 12,
                "health_check": {
                    "protocol": "http",
                    "port": 13,
                    "interval": 1,
                    "timeout": 1,
                    "retries": 1,
                },
            },
            url="/load_balancers/14/actions/update_service",
            method="POST",
        )

        assert action.id == 13
        assert action.progress == 100
        assert action.command == "update_service"

    def test_change_algorithm(
        self, hetzner_client, response_change_algorithm, bound_load_balancer
    ):
        hetzner_client.request.return_value = response_change_algorithm
        algorithm = LoadBalancerAlgorithm(type="round_robin")
        action = bound_load_balancer.change_algorithm(algorithm)
        hetzner_client.request.assert_called_with(
            json={"type": "round_robin"},
            url="/load_balancers/14/actions/change_algorithm",
            method="POST",
        )

        assert action.id == 13
        assert action.progress == 100
        assert action.command == "change_algorithm"

    def test_change_dns_ptr(
        self, hetzner_client, response_change_reverse_dns_entry, bound_load_balancer
    ):
        hetzner_client.request.return_value = response_change_reverse_dns_entry
        action = bound_load_balancer.change_dns_ptr(
            ip="1.2.3.4", dns_ptr="lb1.example.com"
        )
        hetzner_client.request.assert_called_with(
            json={"dns_ptr": "lb1.example.com", "ip": "1.2.3.4"},
            url="/load_balancers/14/actions/change_dns_ptr",
            method="POST",
        )

        assert action.id == 13
        assert action.progress == 100
        assert action.command == "change_dns_ptr"

    def test_change_protection(
        self, hetzner_client, response_change_protection, bound_load_balancer
    ):
        hetzner_client.request.return_value = response_change_protection
        action = bound_load_balancer.change_protection(delete=True)
        hetzner_client.request.assert_called_with(
            json={"delete": True},
            url="/load_balancers/14/actions/change_protection",
            method="POST",
        )

        assert action.id == 13
        assert action.progress == 100
        assert action.command == "change_protection"

    def test_enable_public_interface(
        self, response_enable_public_interface, hetzner_client, bound_load_balancer
    ):
        hetzner_client.request.return_value = response_enable_public_interface
        action = bound_load_balancer.enable_public_interface()
        hetzner_client.request.assert_called_with(
            url="/load_balancers/14/actions/enable_public_interface", method="POST"
        )

        assert action.id == 13
        assert action.progress == 100
        assert action.command == "enable_public_interface"

    def test_disable_public_interface(
        self, response_disable_public_interface, hetzner_client, bound_load_balancer
    ):
        hetzner_client.request.return_value = response_disable_public_interface
        action = bound_load_balancer.disable_public_interface()
        hetzner_client.request.assert_called_with(
            url="/load_balancers/14/actions/disable_public_interface", method="POST"
        )

        assert action.id == 13
        assert action.progress == 100
        assert action.command == "disable_public_interface"

    def test_attach_to_network(
        self,
        response_attach_load_balancer_to_network,
        hetzner_client,
        bound_load_balancer,
    ):
        hetzner_client.request.return_value = response_attach_load_balancer_to_network
        action = bound_load_balancer.attach_to_network(Network(id=1))
        hetzner_client.request.assert_called_with(
            json={"network": 1},
            url="/load_balancers/14/actions/attach_to_network",
            method="POST",
        )

        assert action.id == 13
        assert action.progress == 100
        assert action.command == "attach_to_network"

    def test_detach_from_network(
        self, response_detach_from_network, hetzner_client, bound_load_balancer
    ):
        hetzner_client.request.return_value = response_detach_from_network
        action = bound_load_balancer.detach_from_network(Network(id=1))
        hetzner_client.request.assert_called_with(
            json={"network": 1},
            url="/load_balancers/14/actions/detach_from_network",
            method="POST",
        )

        assert action.id == 13
        assert action.progress == 100
        assert action.command == "detach_from_network"

    def test_change_type(self, hetzner_client, bound_load_balancer, generic_action):
        hetzner_client.request.return_value = generic_action
        action = bound_load_balancer.change_type(LoadBalancerType(name="lb21"))
        hetzner_client.request.assert_called_with(
            url="/load_balancers/14/actions/change_type",
            method="POST",
            json={"load_balancer_type": "lb21"},
        )

        assert action.id == 1
        assert action.progress == 0


class TestLoadBalancerslient:
    @pytest.fixture()
    def load_balancers_client(self):
        return LoadBalancersClient(client=mock.MagicMock())

    def test_get_by_id(self, load_balancers_client, response_load_balancer):
        load_balancers_client._client.request.return_value = response_load_balancer
        bound_load_balancer = load_balancers_client.get_by_id(1)
        load_balancers_client._client.request.assert_called_with(
            url="/load_balancers/1", method="GET"
        )
        assert bound_load_balancer._client is load_balancers_client
        assert bound_load_balancer.id == 4711
        assert bound_load_balancer.name == "Web Frontend"
        assert bound_load_balancer.outgoing_traffic == 123456
        assert bound_load_balancer.ingoing_traffic == 123456
        assert bound_load_balancer.included_traffic == 654321

    @pytest.mark.parametrize(
        "params",
        [
            {
                "name": "load_balancer1",
                "label_selector": "label1",
                "page": 1,
                "per_page": 10,
            },
            {"name": ""},
            {},
        ],
    )
    def test_get_list(
        self, load_balancers_client, response_simple_load_balancers, params
    ):
        load_balancers_client._client.request.return_value = (
            response_simple_load_balancers
        )
        result = load_balancers_client.get_list(**params)
        load_balancers_client._client.request.assert_called_with(
            url="/load_balancers", method="GET", params=params
        )

        bound_load_balancers = result.load_balancers
        assert result.meta is None

        assert len(bound_load_balancers) == 2

        bound_load_balancer1 = bound_load_balancers[0]
        bound_load_balancer2 = bound_load_balancers[1]

        assert bound_load_balancer1._client is load_balancers_client
        assert bound_load_balancer1.id == 4711
        assert bound_load_balancer1.name == "Web Frontend"

        assert bound_load_balancer2._client is load_balancers_client
        assert bound_load_balancer2.id == 4712
        assert bound_load_balancer2.name == "Web Frontend2"

    @pytest.mark.parametrize(
        "params", [{"name": "loadbalancer1", "label_selector": "label1"}, {}]
    )
    def test_get_all(
        self, load_balancers_client, response_simple_load_balancers, params
    ):
        load_balancers_client._client.request.return_value = (
            response_simple_load_balancers
        )
        bound_load_balancers = load_balancers_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})

        load_balancers_client._client.request.assert_called_with(
            url="/load_balancers", method="GET", params=params
        )

        assert len(bound_load_balancers) == 2

        bound_load_balancer1 = bound_load_balancers[0]
        bound_load_balancer2 = bound_load_balancers[1]

        assert bound_load_balancer1._client is load_balancers_client
        assert bound_load_balancer1.id == 4711
        assert bound_load_balancer1.name == "Web Frontend"

        assert bound_load_balancer2._client is load_balancers_client
        assert bound_load_balancer2.id == 4712
        assert bound_load_balancer2.name == "Web Frontend2"

    def test_get_by_name(self, load_balancers_client, response_simple_load_balancers):
        load_balancers_client._client.request.return_value = (
            response_simple_load_balancers
        )
        bound_load_balancer = load_balancers_client.get_by_name("Web Frontend")

        params = {"name": "Web Frontend"}

        load_balancers_client._client.request.assert_called_with(
            url="/load_balancers", method="GET", params=params
        )

        assert bound_load_balancer._client is load_balancers_client
        assert bound_load_balancer.id == 4711
        assert bound_load_balancer.name == "Web Frontend"

    def test_create(self, load_balancers_client, response_create_load_balancer):
        load_balancers_client._client.request.return_value = (
            response_create_load_balancer
        )
        response = load_balancers_client.create(
            "my-balancer",
            load_balancer_type=LoadBalancerType(name="lb11"),
            location=Location(id=1),
        )
        load_balancers_client._client.request.assert_called_with(
            url="/load_balancers",
            method="POST",
            json={"name": "my-balancer", "load_balancer_type": "lb11", "location": 1},
        )

        bound_load_balancer = response.load_balancer

        assert bound_load_balancer._client is load_balancers_client
        assert bound_load_balancer.id == 1
        assert bound_load_balancer.name == "my-balancer"

    @pytest.mark.parametrize(
        "load_balancer",
        [LoadBalancer(id=1), BoundLoadBalancer(mock.MagicMock(), dict(id=1))],
    )
    def test_change_type_with_load_balancer_type_name(
        self, load_balancers_client, load_balancer, generic_action
    ):
        load_balancers_client._client.request.return_value = generic_action
        action = load_balancers_client.change_type(
            load_balancer, LoadBalancerType(name="lb11")
        )
        load_balancers_client._client.request.assert_called_with(
            url="/load_balancers/1/actions/change_type",
            method="POST",
            json={"load_balancer_type": "lb11"},
        )

        assert action.id == 1
        assert action.progress == 0

    @pytest.mark.parametrize(
        "load_balancer",
        [LoadBalancer(id=1), BoundLoadBalancer(mock.MagicMock(), dict(id=1))],
    )
    def test_change_type_with_load_balancer_type_id(
        self, load_balancers_client, load_balancer, generic_action
    ):
        load_balancers_client._client.request.return_value = generic_action
        action = load_balancers_client.change_type(
            load_balancer, LoadBalancerType(id=1)
        )
        load_balancers_client._client.request.assert_called_with(
            url="/load_balancers/1/actions/change_type",
            method="POST",
            json={"load_balancer_type": 1},
        )

        assert action.id == 1
        assert action.progress == 0

    def test_actions_get_by_id(self, load_balancers_client, response_get_actions):
        load_balancers_client._client.request.return_value = {
            "action": response_get_actions["actions"][0]
        }
        action = load_balancers_client.actions.get_by_id(13)

        load_balancers_client._client.request.assert_called_with(
            url="/load_balancers/actions/13", method="GET"
        )

        assert isinstance(action, BoundAction)
        assert action._client == load_balancers_client._client.actions
        assert action.id == 13
        assert action.command == "change_protection"

    def test_actions_get_list(self, load_balancers_client, response_get_actions):
        load_balancers_client._client.request.return_value = response_get_actions
        result = load_balancers_client.actions.get_list()

        load_balancers_client._client.request.assert_called_with(
            url="/load_balancers/actions",
            method="GET",
            params={},
        )

        actions = result.actions
        assert result.meta is None

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == load_balancers_client._client.actions
        assert actions[0].id == 13
        assert actions[0].command == "change_protection"

    def test_actions_get_all(self, load_balancers_client, response_get_actions):
        load_balancers_client._client.request.return_value = response_get_actions
        actions = load_balancers_client.actions.get_all()

        load_balancers_client._client.request.assert_called_with(
            url="/load_balancers/actions",
            method="GET",
            params={"page": 1, "per_page": 50},
        )

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == load_balancers_client._client.actions
        assert actions[0].id == 13
        assert actions[0].command == "change_protection"



from __future__ import annotations


class BaseDomain:
    __slots__ = ()

    @classmethod
    def from_dict(cls, data: dict):  # type: ignore[no-untyped-def]
        """
        Build the domain object from the data dict.
        """
        supported_data = {k: v for k, v in data.items() if k in cls.__slots__}
        return cls(**supported_data)

    def __repr__(self) -> str:
        kwargs = [f"{key}={getattr(self, key)!r}" for key in self.__slots__]  # type: ignore[var-annotated]
        return f"{self.__class__.__qualname__}({', '.join(kwargs)})"


class DomainIdentityMixin:
    __slots__ = ()

    id: int | None
    name: str | None

    @property
    def id_or_name(self) -> int | str:
        """
        Return the first defined value, and fails if none is defined.
        """
        if self.id is not None:
            return self.id
        if self.name is not None:
            return self.name
        raise ValueError("id or name must be set")


class Pagination(BaseDomain):
    __slots__ = (
        "page",
        "per_page",
        "previous_page",
        "next_page",
        "last_page",
        "total_entries",
    )

    def __init__(
        self,
        page: int,
        per_page: int,
        previous_page: int | None = None,
        next_page: int | None = None,
        last_page: int | None = None,
        total_entries: int | None = None,
    ):
        self.page = page
        self.per_page = per_page
        self.previous_page = previous_page
        self.next_page = next_page
        self.last_page = last_page
        self.total_entries = total_entries


class Meta(BaseDomain):
    __slots__ = ("pagination",)

    def __init__(self, pagination: Pagination | None = None):
        self.pagination = pagination

    @classmethod
    def parse_meta(cls, response: dict) -> Meta | None:
        """
        If present, extract the meta details from the response and return a meta object.
        """
        meta = None
        if response and "meta" in response:
            meta = cls()
            try:
                meta.pagination = Pagination(**response["meta"]["pagination"])
            except KeyError:
                pass

        return meta



from __future__ import annotations

from unittest import mock

import pytest

from hcloud.actions import BoundAction
from hcloud.certificates import (
    BoundCertificate,
    Certificate,
    CertificatesClient,
    ManagedCertificateStatus,
)


class TestBoundCertificate:
    @pytest.fixture()
    def bound_certificate(self, hetzner_client):
        return BoundCertificate(client=hetzner_client.certificates, data=dict(id=14))

    @pytest.mark.parametrize("params", [{"page": 1, "per_page": 10}, {}])
    def test_get_actions_list(
        self, hetzner_client, bound_certificate, response_get_actions, params
    ):
        hetzner_client.request.return_value = response_get_actions
        result = bound_certificate.get_actions_list(**params)
        hetzner_client.request.assert_called_with(
            url="/certificates/14/actions", method="GET", params=params
        )

        actions = result.actions
        assert result.meta is None

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == hetzner_client.actions
        assert actions[0].id == 13
        assert actions[0].command == "change_protection"

    def test_get_actions(self, hetzner_client, bound_certificate, response_get_actions):
        hetzner_client.request.return_value = response_get_actions
        actions = bound_certificate.get_actions()

        params = {"page": 1, "per_page": 50}

        hetzner_client.request.assert_called_with(
            url="/certificates/14/actions", method="GET", params=params
        )

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == hetzner_client.actions
        assert actions[0].id == 13
        assert actions[0].command == "change_protection"

    def test_bound_certificate_init(self, certificate_response):
        bound_certificate = BoundCertificate(
            client=mock.MagicMock(), data=certificate_response["certificate"]
        )

        assert bound_certificate.id == 2323
        assert bound_certificate.name == "My Certificate"
        assert bound_certificate.type == "managed"
        assert (
            bound_certificate.fingerprint
            == "03:c7:55:9b:2a:d1:04:17:09:f6:d0:7f:18:34:63:d4:3e:5f"
        )
        assert bound_certificate.certificate == "-----BEGIN CERTIFICATE-----\n..."
        assert len(bound_certificate.domain_names) == 3
        assert bound_certificate.domain_names[0] == "example.com"
        assert bound_certificate.domain_names[1] == "webmail.example.com"
        assert bound_certificate.domain_names[2] == "www.example.com"
        assert isinstance(bound_certificate.status, ManagedCertificateStatus)
        assert bound_certificate.status.issuance == "failed"
        assert bound_certificate.status.renewal == "scheduled"
        assert bound_certificate.status.error.code == "error_code"
        assert bound_certificate.status.error.message == "error message"

    def test_update(
        self, hetzner_client, bound_certificate, response_update_certificate
    ):
        hetzner_client.request.return_value = response_update_certificate
        certificate = bound_certificate.update(name="New name")
        hetzner_client.request.assert_called_with(
            url="/certificates/14", method="PUT", json={"name": "New name"}
        )

        assert certificate.id == 2323
        assert certificate.name == "New name"

    def test_delete(self, hetzner_client, bound_certificate, generic_action):
        hetzner_client.request.return_value = generic_action
        delete_success = bound_certificate.delete()
        hetzner_client.request.assert_called_with(
            url="/certificates/14", method="DELETE"
        )

        assert delete_success is True

    def test_retry_issuance(
        self, hetzner_client, bound_certificate, response_retry_issuance_action
    ):
        hetzner_client.request.return_value = response_retry_issuance_action
        action = bound_certificate.retry_issuance()
        hetzner_client.request.assert_called_with(
            url="/certificates/14/actions/retry", method="POST"
        )

        assert action.id == 14
        assert action.command == "issue_certificate"


class TestCertificatesClient:
    @pytest.fixture()
    def certificates_client(self):
        return CertificatesClient(client=mock.MagicMock())

    def test_get_by_id(self, certificates_client, certificate_response):
        certificates_client._client.request.return_value = certificate_response
        certificate = certificates_client.get_by_id(1)
        certificates_client._client.request.assert_called_with(
            url="/certificates/1", method="GET"
        )
        assert certificate._client is certificates_client
        assert certificate.id == 2323
        assert certificate.name == "My Certificate"

    @pytest.mark.parametrize(
        "params",
        [
            {
                "name": "My Certificate",
                "label_selector": "k==v",
                "page": 1,
                "per_page": 10,
            },
            {"name": ""},
            {},
        ],
    )
    def test_get_list(self, certificates_client, two_certificates_response, params):
        certificates_client._client.request.return_value = two_certificates_response
        result = certificates_client.get_list(**params)
        certificates_client._client.request.assert_called_with(
            url="/certificates", method="GET", params=params
        )

        certificates = result.certificates
        assert len(certificates) == 2

        certificates1 = certificates[0]
        certificates2 = certificates[1]

        assert certificates1._client is certificates_client
        assert certificates1.id == 2323
        assert certificates1.name == "My Certificate"

        assert certificates2._client is certificates_client
        assert certificates2.id == 2324
        assert certificates2.name == "My website cert"

    @pytest.mark.parametrize(
        "params", [{"name": "My Certificate", "label_selector": "label1"}, {}]
    )
    def test_get_all(self, certificates_client, two_certificates_response, params):
        certificates_client._client.request.return_value = two_certificates_response
        certificates = certificates_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})
        certificates_client._client.request.assert_called_with(
            url="/certificates", method="GET", params=params
        )

        assert len(certificates) == 2

        certificates1 = certificates[0]
        certificates2 = certificates[1]

        assert certificates1._client is certificates_client
        assert certificates1.id == 2323
        assert certificates1.name == "My Certificate"

        assert certificates2._client is certificates_client
        assert certificates2.id == 2324
        assert certificates2.name == "My website cert"

    def test_get_by_name(self, certificates_client, one_certificates_response):
        certificates_client._client.request.return_value = one_certificates_response
        certificates = certificates_client.get_by_name("My Certificate")

        params = {"name": "My Certificate"}
        certificates_client._client.request.assert_called_with(
            url="/certificates", method="GET", params=params
        )

        assert certificates._client is certificates_client
        assert certificates.id == 2323
        assert certificates.name == "My Certificate"

    def test_create(self, certificates_client, certificate_response):
        certificates_client._client.request.return_value = certificate_response
        certificate = certificates_client.create(
            name="My Certificate",
            certificate="-----BEGIN CERTIFICATE-----\n...",
            private_key="-----BEGIN PRIVATE KEY-----\n...",
        )
        certificates_client._client.request.assert_called_with(
            url="/certificates",
            method="POST",
            json={
                "name": "My Certificate",
                "certificate": "-----BEGIN CERTIFICATE-----\n...",
                "private_key": "-----BEGIN PRIVATE KEY-----\n...",
                "type": "uploaded",
            },
        )

        assert certificate.id == 2323
        assert certificate.name == "My Certificate"

    def test_create_managed(
        self, certificates_client, create_managed_certificate_response
    ):
        certificates_client._client.request.return_value = (
            create_managed_certificate_response
        )
        create_managed_certificate_rsp = certificates_client.create_managed(
            name="My Certificate", domain_names=["example.com", "*.example.org"]
        )
        certificates_client._client.request.assert_called_with(
            url="/certificates",
            method="POST",
            json={
                "name": "My Certificate",
                "domain_names": ["example.com", "*.example.org"],
                "type": "managed",
            },
        )

        assert create_managed_certificate_rsp.certificate.id == 2323
        assert create_managed_certificate_rsp.certificate.name == "My Certificate"
        assert create_managed_certificate_rsp.action.id == 14
        assert create_managed_certificate_rsp.action.command == "issue_certificate"

    @pytest.mark.parametrize(
        "certificate",
        [Certificate(id=1), BoundCertificate(mock.MagicMock(), dict(id=1))],
    )
    def test_update(
        self, certificates_client, certificate, response_update_certificate
    ):
        certificates_client._client.request.return_value = response_update_certificate
        certificate = certificates_client.update(certificate, name="New name")
        certificates_client._client.request.assert_called_with(
            url="/certificates/1", method="PUT", json={"name": "New name"}
        )

        assert certificate.id == 2323
        assert certificate.name == "New name"

    @pytest.mark.parametrize(
        "certificate",
        [Certificate(id=1), BoundCertificate(mock.MagicMock(), dict(id=1))],
    )
    def test_delete(self, certificates_client, certificate, generic_action):
        certificates_client._client.request.return_value = generic_action
        delete_success = certificates_client.delete(certificate)
        certificates_client._client.request.assert_called_with(
            url="/certificates/1", method="DELETE"
        )

        assert delete_success is True

    @pytest.mark.parametrize(
        "certificate",
        [Certificate(id=1), BoundCertificate(mock.MagicMock(), dict(id=1))],
    )
    def test_retry_issuance(
        self, certificates_client, certificate, response_retry_issuance_action
    ):
        certificates_client._client.request.return_value = (
            response_retry_issuance_action
        )
        action = certificates_client.retry_issuance(certificate)
        certificates_client._client.request.assert_called_with(
            url="/certificates/1/actions/retry", method="POST"
        )

        assert action.id == 14
        assert action.command == "issue_certificate"

    def test_actions_get_by_id(self, certificates_client, response_get_actions):
        certificates_client._client.request.return_value = {
            "action": response_get_actions["actions"][0]
        }
        action = certificates_client.actions.get_by_id(13)

        certificates_client._client.request.assert_called_with(
            url="/certificates/actions/13", method="GET"
        )

        assert isinstance(action, BoundAction)
        assert action._client == certificates_client._client.actions
        assert action.id == 13
        assert action.command == "change_protection"

    def test_actions_get_list(self, certificates_client, response_get_actions):
        certificates_client._client.request.return_value = response_get_actions
        result = certificates_client.actions.get_list()

        certificates_client._client.request.assert_called_with(
            url="/certificates/actions",
            method="GET",
            params={},
        )

        actions = result.actions
        assert result.meta is None

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == certificates_client._client.actions
        assert actions[0].id == 13
        assert actions[0].command == "change_protection"

    def test_actions_get_all(self, certificates_client, response_get_actions):
        certificates_client._client.request.return_value = response_get_actions
        actions = certificates_client.actions.get_all()

        certificates_client._client.request.assert_called_with(
            url="/certificates/actions",
            method="GET",
            params={"page": 1, "per_page": 50},
        )

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == certificates_client._client.actions
        assert actions[0].id == 13
        assert actions[0].command == "change_protection"



from __future__ import annotations

import pytest


@pytest.fixture()
def response_create_placement_group():
    return {
        "placement_group": {
            "created": "2019-01-08T12:10:00+00:00",
            "id": 897,
            "labels": {"key": "value"},
            "name": "my Placement Group",
            "servers": [],
            "type": "spread",
        }
    }


@pytest.fixture()
def one_placement_group_response():
    return {
        "placement_groups": [
            {
                "created": "2019-01-08T12:10:00+00:00",
                "id": 897,
                "labels": {"key": "value"},
                "name": "my Placement Group",
                "servers": [4711, 4712],
                "type": "spread",
            }
        ]
    }


@pytest.fixture()
def two_placement_groups_response():
    return {
        "placement_groups": [
            {
                "created": "2019-01-08T12:10:00+00:00",
                "id": 897,
                "labels": {"key": "value"},
                "name": "my Placement Group",
                "servers": [4711, 4712],
                "type": "spread",
            },
            {
                "created": "2019-01-08T12:10:00+00:00",
                "id": 898,
                "labels": {"key": "value"},
                "name": "my Placement Group",
                "servers": [4713, 4714, 4715],
                "type": "spread",
            },
        ]
    }


@pytest.fixture()
def placement_group_response():
    return {
        "placement_group": {
            "created": "2019-01-08T12:10:00+00:00",
            "id": 897,
            "labels": {"key": "value"},
            "name": "my Placement Group",
            "servers": [4711, 4712],
            "type": "spread",
        }
    }



from __future__ import annotations

import json
from unittest.mock import MagicMock

import pytest
import requests

from hcloud import APIException, Client


class TestHetznerClient:
    @pytest.fixture()
    def client(self):
        Client._version = "0.0.0"
        client = Client(token="project_token")

        client._requests_session = MagicMock()
        return client

    @pytest.fixture()
    def response(self):
        response = requests.Response()
        response.status_code = 200
        response._content = json.dumps({"result": "data"}).encode("utf-8")
        return response

    @pytest.fixture()
    def fail_response(self, response):
        response.status_code = 422
        error = {
            "code": "invalid_input",
            "message": "invalid input in field 'broken_field': is too long",
            "details": {
                "fields": [{"name": "broken_field", "messages": ["is too long"]}]
            },
        }
        response._content = json.dumps({"error": error}).encode("utf-8")
        return response

    @pytest.fixture()
    def rate_limit_response(self, response):
        response.status_code = 422
        error = {
            "code": "rate_limit_exceeded",
            "message": "limit of 10 requests per hour reached",
            "details": {},
        }
        response._content = json.dumps({"error": error}).encode("utf-8")
        return response

    def test__get_user_agent(self, client):
        user_agent = client._get_user_agent()
        assert user_agent == "hcloud-python/0.0.0"

    def test__get_user_agent_with_application_name(self, client):
        client = Client(token="project_token", application_name="my-app")
        user_agent = client._get_user_agent()
        assert user_agent == "my-app hcloud-python/0.0.0"

    def test__get_user_agent_with_application_name_and_version(self, client):
        client = Client(
            token="project_token",
            application_name="my-app",
            application_version="1.0.0",
        )
        user_agent = client._get_user_agent()
        assert user_agent == "my-app/1.0.0 hcloud-python/0.0.0"

    def test__get_headers(self, client):
        headers = client._get_headers()
        assert headers == {
            "User-Agent": "hcloud-python/0.0.0",
            "Authorization": "Bearer project_token",
        }

    def test_request_library_mocked(self, client):
        response = client.request("POST", "url", params={"1": 2})
        assert response.__class__.__name__ == "MagicMock"

    def test_request_ok(self, client, response):
        client._requests_session.request.return_value = response
        response = client.request(
            "POST", "/servers", params={"argument": "value"}, timeout=2
        )
        client._requests_session.request.assert_called_once_with(
            method="POST",
            url="https://api.hetzner.cloud/v1/servers",
            headers={
                "User-Agent": "hcloud-python/0.0.0",
                "Authorization": "Bearer project_token",
            },
            params={"argument": "value"},
            timeout=2,
        )
        assert response == {"result": "data"}

    def test_request_fails(self, client, fail_response):
        client._requests_session.request.return_value = fail_response
        with pytest.raises(APIException) as exception_info:
            client.request(
                "POST", "http://url.com", params={"argument": "value"}, timeout=2
            )
        error = exception_info.value
        assert error.code == "invalid_input"
        assert error.message == "invalid input in field 'broken_field': is too long"
        assert error.details["fields"][0]["name"] == "broken_field"

    def test_request_500(self, client, fail_response):
        fail_response.status_code = 500
        fail_response.reason = "Internal Server Error"
        fail_response._content = "Internal Server Error"
        client._requests_session.request.return_value = fail_response
        with pytest.raises(APIException) as exception_info:
            client.request(
                "POST", "http://url.com", params={"argument": "value"}, timeout=2
            )
        error = exception_info.value
        assert error.code == 500
        assert error.message == "Internal Server Error"
        assert error.details["content"] == "Internal Server Error"

    def test_request_broken_json_200(self, client, response):
        content = b"{'key': 'value'"
        response.reason = "OK"
        response._content = content
        client._requests_session.request.return_value = response
        with pytest.raises(APIException) as exception_info:
            client.request(
                "POST", "http://url.com", params={"argument": "value"}, timeout=2
            )
        error = exception_info.value
        assert error.code == 200
        assert error.message == "OK"
        assert error.details["content"] == content

    def test_request_empty_content_200(self, client, response):
        content = ""
        response.reason = "OK"
        response._content = content
        client._requests_session.request.return_value = response
        response = client.request(
            "POST", "http://url.com", params={"argument": "value"}, timeout=2
        )
        assert response == ""

    def test_request_500_empty_content(self, client, fail_response):
        fail_response.status_code = 500
        fail_response.reason = "Internal Server Error"
        fail_response._content = ""
        client._requests_session.request.return_value = fail_response
        with pytest.raises(APIException) as exception_info:
            client.request(
                "POST", "http://url.com", params={"argument": "value"}, timeout=2
            )
        error = exception_info.value
        assert error.code == 500
        assert error.message == "Internal Server Error"
        assert error.details["content"] == ""
        assert str(error) == "Internal Server Error"

    def test_request_limit(self, client, rate_limit_response):
        client._retry_wait_time = 0
        client._requests_session.request.return_value = rate_limit_response
        with pytest.raises(APIException) as exception_info:
            client.request(
                "POST", "http://url.com", params={"argument": "value"}, timeout=2
            )
        error = exception_info.value
        assert client._requests_session.request.call_count == 5
        assert error.code == "rate_limit_exceeded"
        assert error.message == "limit of 10 requests per hour reached"

    def test_request_limit_then_success(self, client, rate_limit_response):
        client._retry_wait_time = 0
        response = requests.Response()
        response.status_code = 200
        response._content = json.dumps({"result": "data"}).encode("utf-8")
        client._requests_session.request.side_effect = [rate_limit_response, response]

        client.request(
            "POST", "http://url.com", params={"argument": "value"}, timeout=2
        )
        assert client._requests_session.request.call_count == 2



from __future__ import annotations

import pytest


@pytest.fixture()
def location_response():
    return {
        "location": {
            "id": 1,
            "name": "fsn1",
            "description": "Falkenstein DC Park 1",
            "country": "DE",
            "city": "Falkenstein",
            "latitude": 50.47612,
            "longitude": 12.370071,
            "network_zone": "eu-central",
        }
    }


@pytest.fixture()
def two_locations_response():
    return {
        "locations": [
            {
                "id": 1,
                "name": "fsn1",
                "description": "Falkenstein DC Park 1",
                "country": "DE",
                "city": "Falkenstein",
                "latitude": 50.47612,
                "longitude": 12.370071,
                "network_zone": "eu-central",
            },
            {
                "id": 2,
                "name": "nbg1",
                "description": "Nuremberg DC Park 1",
                "country": "DE",
                "city": "Nuremberg",
                "latitude": 49.452102,
                "longitude": 11.076665,
                "network_zone": "eu-central",
            },
        ]
    }


@pytest.fixture()
def one_locations_response():
    return {
        "locations": [
            {
                "id": 1,
                "name": "fsn1",
                "description": "Falkenstein DC Park 1",
                "country": "DE",
                "city": "Falkenstein",
                "latitude": 50.47612,
                "longitude": 12.370071,
                "network_zone": "eu-central",
            }
        ]
    }



from __future__ import annotations

import pytest


@pytest.fixture()
def iso_response():
    return {
        "iso": {
            "id": 4711,
            "name": "FreeBSD-11.0-RELEASE-amd64-dvd1",
            "description": "FreeBSD 11.0 x64",
            "type": "public",
            "architecture": "x86",
            "deprecated": "2018-02-28T00:00:00+00:00",
            "deprecation": {
                "announced": "2018-01-28T00:00:00+00:00",
                "unavailable_after": "2018-02-28T00:00:00+00:00",
            },
        }
    }


@pytest.fixture()
def two_isos_response():
    return {
        "isos": [
            {
                "id": 4711,
                "name": "FreeBSD-11.0-RELEASE-amd64-dvd1",
                "description": "FreeBSD 11.0 x64",
                "type": "public",
                "architecture": "x86",
                "deprecated": "2018-02-28T00:00:00+00:00",
                "deprecation": {
                    "announced": "2018-01-28T00:00:00+00:00",
                    "unavailable_after": "2018-02-28T00:00:00+00:00",
                },
            },
            {
                "id": 4712,
                "name": "FreeBSD-11.0-RELEASE-amd64-dvd1",
                "description": "FreeBSD 11.0 x64",
                "type": "public",
                "architecture": "x86",
                "deprecated": None,
            },
        ]
    }


@pytest.fixture()
def one_isos_response():
    return {
        "isos": [
            {
                "id": 4711,
                "name": "FreeBSD-11.0-RELEASE-amd64-dvd1",
                "description": "FreeBSD 11.0 x64",
                "type": "public",
                "architecture": "x86",
                "deprecated": "2018-02-28T00:00:00+00:00",
                "deprecation": {
                    "announced": "2018-01-28T00:00:00+00:00",
                    "unavailable_after": "2018-02-28T00:00:00+00:00",
                },
            }
        ]
    }



from __future__ import annotations

import pytest


@pytest.fixture()
def response_create_firewall():
    return {
        "firewall": {
            "id": 38,
            "name": "Corporate Intranet Protection",
            "labels": {},
            "created": "2016-01-30T23:50:00+00:00",
            "rules": [
                {
                    "direction": "in",
                    "source_ips": [
                        "28.239.13.1/32",
                        "28.239.14.0/24",
                        "ff21:1eac:9a3b:ee58:5ca:990c:8bc9:c03b/128",
                    ],
                    "destination_ips": [],
                    "protocol": "tcp",
                    "port": "80",
                    "description": None,
                },
                {
                    "direction": "out",
                    "source_ips": [],
                    "destination_ips": [
                        "28.239.13.1/32",
                        "28.239.14.0/24",
                        "ff21:1eac:9a3b:ee58:5ca:990c:8bc9:c03b/128",
                    ],
                    "protocol": "tcp",
                    "port": "80",
                    "description": "allow http out",
                },
            ],
            "applied_to": [
                {"server": {"id": 42}, "type": "server"},
                {
                    "type": "label_selector",
                    "label_selector": {"selector": "key==value"},
                },
            ],
        },
        "actions": [
            {
                "command": "set_firewall_rules",
                "error": {"code": "action_failed", "message": "Action failed"},
                "finished": "2016-01-30T23:56:00+00:00",
                "id": 13,
                "progress": 100,
                "resources": [{"id": 38, "type": "firewall"}],
                "started": "2016-01-30T23:55:00+00:00",
                "status": "success",
            },
            {
                "command": "apply_firewall",
                "error": {"code": "action_failed", "message": "Action failed"},
                "finished": "2016-01-30T23:56:00+00:00",
                "id": 14,
                "progress": 100,
                "resources": [
                    {"id": 42, "type": "server"},
                    {"id": 38, "type": "firewall"},
                ],
                "started": "2016-01-30T23:55:00+00:00",
                "status": "success",
            },
        ],
    }


@pytest.fixture()
def firewall_response():
    return {
        "firewall": {
            "id": 38,
            "name": "Corporate Intranet Protection",
            "labels": {},
            "created": "2016-01-30T23:50:00+00:00",
            "rules": [
                {
                    "direction": "in",
                    "source_ips": [
                        "28.239.13.1/32",
                        "28.239.14.0/24",
                        "ff21:1eac:9a3b:ee58:5ca:990c:8bc9:c03b/128",
                    ],
                    "destination_ips": [],
                    "protocol": "tcp",
                    "port": "80",
                    "description": "allow http in",
                },
                {
                    "direction": "out",
                    "source_ips": [],
                    "destination_ips": [
                        "28.239.13.1/32",
                        "28.239.14.0/24",
                        "ff21:1eac:9a3b:ee58:5ca:990c:8bc9:c03b/128",
                    ],
                    "protocol": "tcp",
                    "port": "80",
                    "description": "allow http out",
                },
            ],
            "applied_to": [
                {"server": {"id": 42}, "type": "server"},
                {
                    "type": "label_selector",
                    "label_selector": {"selector": "key==value"},
                },
            ],
        }
    }


@pytest.fixture()
def two_firewalls_response():
    return {
        "firewalls": [
            {
                "id": 38,
                "name": "Corporate Intranet Protection",
                "labels": {},
                "created": "2016-01-30T23:50:00+00:00",
                "rules": [
                    {
                        "direction": "in",
                        "source_ips": [
                            "28.239.13.1/32",
                            "28.239.14.0/24",
                            "ff21:1eac:9a3b:ee58:5ca:990c:8bc9:c03b/128",
                        ],
                        "destination_ips": [],
                        "protocol": "tcp",
                        "port": "80",
                        "description": "allow http in",
                    }
                ],
                "applied_to": [{"server": {"id": 42}, "type": "server"}],
            },
            {
                "id": 39,
                "name": "Corporate Extranet Protection",
                "labels": {},
                "created": "2016-01-30T23:50:00+00:00",
                "rules": [
                    {
                        "direction": "in",
                        "destination_ips": [],
                        "source_ips": [
                            "28.239.13.1/32",
                            "28.239.14.0/24",
                            "ff21:1eac:9a3b:ee58:5ca:990c:8bc9:c03b/128",
                        ],
                        "protocol": "tcp",
                        "port": "443",
                        "description": "allow https in",
                    }
                ],
                "applied_to": [{"server": {"id": 42}, "type": "server"}],
            },
        ]
    }


@pytest.fixture()
def one_firewalls_response():
    return {
        "firewalls": [
            {
                "id": 38,
                "name": "Corporate Intranet Protection",
                "labels": {},
                "created": "2016-01-30T23:50:00+00:00",
                "rules": [
                    {
                        "direction": "in",
                        "destination_ips": [],
                        "source_ips": [
                            "28.239.13.1/32",
                            "28.239.14.0/24",
                            "ff21:1eac:9a3b:ee58:5ca:990c:8bc9:c03b/128",
                        ],
                        "protocol": "tcp",
                        "port": "80",
                        "description": "allow http in",
                    }
                ],
                "applied_to": [{"server": {"id": 42}, "type": "server"}],
            }
        ]
    }


@pytest.fixture()
def response_update_firewall():
    return {
        "firewall": {
            "id": 38,
            "name": "New Corporate Intranet Protection",
            "labels": {},
            "created": "2016-01-30T23:50:00+00:00",
            "rules": [
                {
                    "direction": "in",
                    "source_ips": [
                        "28.239.13.1/32",
                        "28.239.14.0/24",
                        "ff21:1eac:9a3b:ee58:5ca:990c:8bc9:c03b/128",
                    ],
                    "destination_ips": [],
                    "protocol": "tcp",
                    "port": "80",
                    "description": "allow http in",
                }
            ],
            "applied_to": [{"server": {"id": 42}, "type": "server"}],
        }
    }


@pytest.fixture()
def response_get_actions():
    return {
        "actions": [
            {
                "id": 13,
                "command": "set_firewall_rules",
                "status": "success",
                "progress": 100,
                "started": "2016-01-30T23:55:00+00:00",
                "finished": "2016-01-30T23:56:00+00:00",
                "resources": [{"id": 42, "type": "firewall"}],
                "error": {"code": "action_failed", "message": "Action failed"},
            }
        ]
    }


@pytest.fixture()
def response_set_rules():
    return {
        "actions": [
            {
                "id": 13,
                "command": "set_firewall_rules",
                "status": "success",
                "progress": 100,
                "started": "2016-01-30T23:55:00+00:00",
                "finished": "2016-01-30T23:56:00+00:00",
                "resources": [{"id": 38, "type": "firewall"}],
                "error": {"code": "action_failed", "message": "Action failed"},
            },
            {
                "id": 14,
                "command": "apply_firewall",
                "status": "success",
                "progress": 100,
                "started": "2016-01-30T23:55:00+00:00",
                "finished": "2016-01-30T23:56:00+00:00",
                "resources": [
                    {"id": 38, "type": "firewall"},
                    {"id": 42, "type": "server"},
                ],
                "error": {"code": "action_failed", "message": "Action failed"},
            },
        ]
    }



from __future__ import annotations

from unittest import mock

import pytest

from hcloud.actions import BoundAction
from hcloud.floating_ips import BoundFloatingIP, FloatingIP, FloatingIPsClient
from hcloud.locations import BoundLocation, Location
from hcloud.servers import BoundServer, Server


class TestBoundFloatingIP:
    @pytest.fixture()
    def bound_floating_ip(self, hetzner_client):
        return BoundFloatingIP(client=hetzner_client.floating_ips, data=dict(id=14))

    def test_bound_floating_ip_init(self, floating_ip_response):
        bound_floating_ip = BoundFloatingIP(
            client=mock.MagicMock(), data=floating_ip_response["floating_ip"]
        )

        assert bound_floating_ip.id == 4711
        assert bound_floating_ip.description == "Web Frontend"
        assert bound_floating_ip.name == "Web Frontend"
        assert bound_floating_ip.ip == "131.232.99.1"
        assert bound_floating_ip.type == "ipv4"
        assert bound_floating_ip.protection == {"delete": False}
        assert bound_floating_ip.labels == {}
        assert bound_floating_ip.blocked is False

        assert isinstance(bound_floating_ip.server, BoundServer)
        assert bound_floating_ip.server.id == 42

        assert isinstance(bound_floating_ip.home_location, BoundLocation)
        assert bound_floating_ip.home_location.id == 1
        assert bound_floating_ip.home_location.name == "fsn1"
        assert bound_floating_ip.home_location.description == "Falkenstein DC Park 1"
        assert bound_floating_ip.home_location.country == "DE"
        assert bound_floating_ip.home_location.city == "Falkenstein"
        assert bound_floating_ip.home_location.latitude == 50.47612
        assert bound_floating_ip.home_location.longitude == 12.370071

    def test_get_actions(self, hetzner_client, bound_floating_ip, response_get_actions):
        hetzner_client.request.return_value = response_get_actions
        actions = bound_floating_ip.get_actions(sort="id")
        hetzner_client.request.assert_called_with(
            url="/floating_ips/14/actions",
            method="GET",
            params={"sort": "id", "page": 1, "per_page": 50},
        )

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == hetzner_client.actions
        assert actions[0].id == 13
        assert actions[0].command == "assign_floating_ip"

    def test_update(
        self, hetzner_client, bound_floating_ip, response_update_floating_ip
    ):
        hetzner_client.request.return_value = response_update_floating_ip
        floating_ip = bound_floating_ip.update(
            description="New description", name="New name"
        )
        hetzner_client.request.assert_called_with(
            url="/floating_ips/14",
            method="PUT",
            json={"description": "New description", "name": "New name"},
        )

        assert floating_ip.id == 4711
        assert floating_ip.description == "New description"
        assert floating_ip.name == "New name"

    def test_delete(self, hetzner_client, bound_floating_ip, generic_action):
        hetzner_client.request.return_value = generic_action
        delete_success = bound_floating_ip.delete()
        hetzner_client.request.assert_called_with(
            url="/floating_ips/14", method="DELETE"
        )

        assert delete_success is True

    def test_change_protection(self, hetzner_client, bound_floating_ip, generic_action):
        hetzner_client.request.return_value = generic_action
        action = bound_floating_ip.change_protection(True)
        hetzner_client.request.assert_called_with(
            url="/floating_ips/14/actions/change_protection",
            method="POST",
            json={"delete": True},
        )

        assert action.id == 1
        assert action.progress == 0

    @pytest.mark.parametrize(
        "server", (Server(id=1), BoundServer(mock.MagicMock(), dict(id=1)))
    )
    def test_assign(self, hetzner_client, bound_floating_ip, server, generic_action):
        hetzner_client.request.return_value = generic_action
        action = bound_floating_ip.assign(server)
        hetzner_client.request.assert_called_with(
            url="/floating_ips/14/actions/assign", method="POST", json={"server": 1}
        )
        assert action.id == 1
        assert action.progress == 0

    def test_unassign(self, hetzner_client, bound_floating_ip, generic_action):
        hetzner_client.request.return_value = generic_action
        action = bound_floating_ip.unassign()
        hetzner_client.request.assert_called_with(
            url="/floating_ips/14/actions/unassign", method="POST"
        )
        assert action.id == 1
        assert action.progress == 0

    def test_change_dns_ptr(self, hetzner_client, bound_floating_ip, generic_action):
        hetzner_client.request.return_value = generic_action
        action = bound_floating_ip.change_dns_ptr("1.2.3.4", "server02.example.com")
        hetzner_client.request.assert_called_with(
            url="/floating_ips/14/actions/change_dns_ptr",
            method="POST",
            json={"ip": "1.2.3.4", "dns_ptr": "server02.example.com"},
        )
        assert action.id == 1
        assert action.progress == 0


class TestFloatingIPsClient:
    @pytest.fixture()
    def floating_ips_client(self):
        return FloatingIPsClient(client=mock.MagicMock())

    def test_get_by_id(self, floating_ips_client, floating_ip_response):
        floating_ips_client._client.request.return_value = floating_ip_response
        bound_floating_ip = floating_ips_client.get_by_id(1)
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips/1", method="GET"
        )
        assert bound_floating_ip._client is floating_ips_client
        assert bound_floating_ip.id == 4711
        assert bound_floating_ip.description == "Web Frontend"

    def test_get_by_name(self, floating_ips_client, one_floating_ips_response):
        floating_ips_client._client.request.return_value = one_floating_ips_response
        bound_floating_ip = floating_ips_client.get_by_name("Web Frontend")
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips", method="GET", params={"name": "Web Frontend"}
        )
        assert bound_floating_ip._client is floating_ips_client
        assert bound_floating_ip.id == 4711
        assert bound_floating_ip.name == "Web Frontend"
        assert bound_floating_ip.description == "Web Frontend"

    @pytest.mark.parametrize(
        "params",
        [{"label_selector": "label1", "page": 1, "per_page": 10}, {"name": ""}, {}],
    )
    def test_get_list(self, floating_ips_client, two_floating_ips_response, params):
        floating_ips_client._client.request.return_value = two_floating_ips_response
        result = floating_ips_client.get_list(**params)
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips", method="GET", params=params
        )

        bound_floating_ips = result.floating_ips
        assert result.meta is None

        assert len(bound_floating_ips) == 2

        bound_floating_ip1 = bound_floating_ips[0]
        bound_floating_ip2 = bound_floating_ips[1]

        assert bound_floating_ip1._client is floating_ips_client
        assert bound_floating_ip1.id == 4711
        assert bound_floating_ip1.description == "Web Frontend"

        assert bound_floating_ip2._client is floating_ips_client
        assert bound_floating_ip2.id == 4712
        assert bound_floating_ip2.description == "Web Backend"

    @pytest.mark.parametrize("params", [{"label_selector": "label1"}, {}])
    def test_get_all(self, floating_ips_client, two_floating_ips_response, params):
        floating_ips_client._client.request.return_value = two_floating_ips_response
        bound_floating_ips = floating_ips_client.get_all(**params)

        params.update({"page": 1, "per_page": 50})

        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips", method="GET", params=params
        )

        assert len(bound_floating_ips) == 2

        bound_floating_ip1 = bound_floating_ips[0]
        bound_floating_ip2 = bound_floating_ips[1]

        assert bound_floating_ip1._client is floating_ips_client
        assert bound_floating_ip1.id == 4711
        assert bound_floating_ip1.description == "Web Frontend"

        assert bound_floating_ip2._client is floating_ips_client
        assert bound_floating_ip2.id == 4712
        assert bound_floating_ip2.description == "Web Backend"

    def test_create_with_location(self, floating_ips_client, floating_ip_response):
        floating_ips_client._client.request.return_value = floating_ip_response
        response = floating_ips_client.create(
            "ipv6", "Web Frontend", home_location=Location(name="location")
        )
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips",
            method="POST",
            json={
                "description": "Web Frontend",
                "type": "ipv6",
                "home_location": "location",
            },
        )

        bound_floating_ip = response.floating_ip
        action = response.action

        assert bound_floating_ip._client is floating_ips_client
        assert bound_floating_ip.id == 4711
        assert bound_floating_ip.description == "Web Frontend"
        assert action is None

    @pytest.mark.parametrize(
        "server", [Server(id=1), BoundServer(mock.MagicMock(), dict(id=1))]
    )
    def test_create_with_server(
        self, floating_ips_client, server, floating_ip_create_response
    ):
        floating_ips_client._client.request.return_value = floating_ip_create_response
        response = floating_ips_client.create(
            type="ipv6", description="Web Frontend", server=server
        )
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips",
            method="POST",
            json={"description": "Web Frontend", "type": "ipv6", "server": 1},
        )
        bound_floating_ip = response.floating_ip
        action = response.action

        assert bound_floating_ip._client is floating_ips_client
        assert bound_floating_ip.id == 4711
        assert bound_floating_ip.description == "Web Frontend"
        assert action.id == 13

    @pytest.mark.parametrize(
        "server", [Server(id=1), BoundServer(mock.MagicMock(), dict(id=1))]
    )
    def test_create_with_name(
        self, floating_ips_client, server, floating_ip_create_response
    ):
        floating_ips_client._client.request.return_value = floating_ip_create_response
        response = floating_ips_client.create(
            type="ipv6", description="Web Frontend", name="Web Frontend"
        )
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips",
            method="POST",
            json={
                "description": "Web Frontend",
                "type": "ipv6",
                "name": "Web Frontend",
            },
        )
        bound_floating_ip = response.floating_ip
        action = response.action

        assert bound_floating_ip._client is floating_ips_client
        assert bound_floating_ip.id == 4711
        assert bound_floating_ip.description == "Web Frontend"
        assert bound_floating_ip.name == "Web Frontend"
        assert action.id == 13

    @pytest.mark.parametrize(
        "floating_ip", [FloatingIP(id=1), BoundFloatingIP(mock.MagicMock(), dict(id=1))]
    )
    def test_get_actions(self, floating_ips_client, floating_ip, response_get_actions):
        floating_ips_client._client.request.return_value = response_get_actions
        actions = floating_ips_client.get_actions(floating_ip)
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips/1/actions",
            method="GET",
            params={"page": 1, "per_page": 50},
        )

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)

        assert actions[0]._client == floating_ips_client._client.actions
        assert actions[0].id == 13
        assert actions[0].command == "assign_floating_ip"

    @pytest.mark.parametrize(
        "floating_ip", [FloatingIP(id=1), BoundFloatingIP(mock.MagicMock(), dict(id=1))]
    )
    def test_update(
        self, floating_ips_client, floating_ip, response_update_floating_ip
    ):
        floating_ips_client._client.request.return_value = response_update_floating_ip
        floating_ip = floating_ips_client.update(
            floating_ip, description="New description", name="New name"
        )
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips/1",
            method="PUT",
            json={"description": "New description", "name": "New name"},
        )

        assert floating_ip.id == 4711
        assert floating_ip.description == "New description"
        assert floating_ip.name == "New name"

    @pytest.mark.parametrize(
        "floating_ip", [FloatingIP(id=1), BoundFloatingIP(mock.MagicMock(), dict(id=1))]
    )
    def test_change_protection(self, floating_ips_client, floating_ip, generic_action):
        floating_ips_client._client.request.return_value = generic_action
        action = floating_ips_client.change_protection(floating_ip, True)
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips/1/actions/change_protection",
            method="POST",
            json={"delete": True},
        )

        assert action.id == 1
        assert action.progress == 0

    @pytest.mark.parametrize(
        "floating_ip", [FloatingIP(id=1), BoundFloatingIP(mock.MagicMock(), dict(id=1))]
    )
    def test_delete(self, floating_ips_client, floating_ip, generic_action):
        floating_ips_client._client.request.return_value = generic_action
        delete_success = floating_ips_client.delete(floating_ip)
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips/1", method="DELETE"
        )

        assert delete_success is True

    @pytest.mark.parametrize(
        "server,floating_ip",
        [
            (Server(id=1), FloatingIP(id=12)),
            (
                BoundServer(mock.MagicMock(), dict(id=1)),
                BoundFloatingIP(mock.MagicMock(), dict(id=12)),
            ),
        ],
    )
    def test_assign(self, floating_ips_client, server, floating_ip, generic_action):
        floating_ips_client._client.request.return_value = generic_action
        action = floating_ips_client.assign(floating_ip, server)
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips/12/actions/assign", method="POST", json={"server": 1}
        )
        assert action.id == 1
        assert action.progress == 0

    @pytest.mark.parametrize(
        "floating_ip",
        [FloatingIP(id=12), BoundFloatingIP(mock.MagicMock(), dict(id=12))],
    )
    def test_unassign(self, floating_ips_client, floating_ip, generic_action):
        floating_ips_client._client.request.return_value = generic_action
        action = floating_ips_client.unassign(floating_ip)
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips/12/actions/unassign", method="POST"
        )
        assert action.id == 1
        assert action.progress == 0

    @pytest.mark.parametrize(
        "floating_ip",
        [FloatingIP(id=12), BoundFloatingIP(mock.MagicMock(), dict(id=12))],
    )
    def test_change_dns_ptr(self, floating_ips_client, floating_ip, generic_action):
        floating_ips_client._client.request.return_value = generic_action
        action = floating_ips_client.change_dns_ptr(
            floating_ip, "1.2.3.4", "server02.example.com"
        )
        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips/12/actions/change_dns_ptr",
            method="POST",
            json={"ip": "1.2.3.4", "dns_ptr": "server02.example.com"},
        )
        assert action.id == 1
        assert action.progress == 0

    def test_actions_get_by_id(self, floating_ips_client, response_get_actions):
        floating_ips_client._client.request.return_value = {
            "action": response_get_actions["actions"][0]
        }
        action = floating_ips_client.actions.get_by_id(13)

        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips/actions/13", method="GET"
        )

        assert isinstance(action, BoundAction)
        assert action._client == floating_ips_client._client.actions
        assert action.id == 13
        assert action.command == "assign_floating_ip"

    def test_actions_get_list(self, floating_ips_client, response_get_actions):
        floating_ips_client._client.request.return_value = response_get_actions
        result = floating_ips_client.actions.get_list()

        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips/actions",
            method="GET",
            params={},
        )

        actions = result.actions
        assert result.meta is None

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == floating_ips_client._client.actions
        assert actions[0].id == 13
        assert actions[0].command == "assign_floating_ip"

    def test_actions_get_all(self, floating_ips_client, response_get_actions):
        floating_ips_client._client.request.return_value = response_get_actions
        actions = floating_ips_client.actions.get_all()

        floating_ips_client._client.request.assert_called_with(
            url="/floating_ips/actions",
            method="GET",
            params={"page": 1, "per_page": 50},
        )

        assert len(actions) == 1
        assert isinstance(actions[0], BoundAction)
        assert actions[0]._client == floating_ips_client._client.actions
        assert actions[0].id == 13
        assert actions[0].command == "assign_floating_ip"



from __future__ import annotations

import time
from typing import NoReturn

import requests

from .__version__ import VERSION
from ._exceptions import APIException
from .actions import ActionsClient
from .certificates import CertificatesClient
from .datacenters import DatacentersClient
from .firewalls import FirewallsClient
from .floating_ips import FloatingIPsClient
from .images import ImagesClient
from .isos import IsosClient
from .load_balancer_types import LoadBalancerTypesClient
from .load_balancers import LoadBalancersClient
from .locations import LocationsClient
from .networks import NetworksClient
from .placement_groups import PlacementGroupsClient
from .primary_ips import PrimaryIPsClient
from .server_types import ServerTypesClient
from .servers import ServersClient
from .ssh_keys import SSHKeysClient
from .volumes import VolumesClient


class Client:
    """Base Client for accessing the Hetzner Cloud API"""

    _version = VERSION
    _retry_wait_time = 0.5
    __user_agent_prefix = "hcloud-python"

    def __init__(
        self,
        token: str,
        api_endpoint: str = "https://api.hetzner.cloud/v1",
        application_name: str | None = None,
        application_version: str | None = None,
        poll_interval: int = 1,
        timeout: float | tuple[float, float] | None = None,
    ):
        """Create a new Client instance

        :param token: Hetzner Cloud API token
        :param api_endpoint: Hetzner Cloud API endpoint
        :param application_name: Your application name
        :param application_version: Your application _version
        :param poll_interval: Interval for polling information from Hetzner Cloud API in seconds
        :param timeout: Requests timeout in seconds
        """
        self.token = token
        self._api_endpoint = api_endpoint
        self._application_name = application_name
        self._application_version = application_version
        self._requests_session = requests.Session()
        self._requests_timeout = timeout
        self.poll_interval = poll_interval

        self.datacenters = DatacentersClient(self)
        """DatacentersClient Instance

        :type: :class:`DatacentersClient <hcloud.datacenters.client.DatacentersClient>`
        """
        self.locations = LocationsClient(self)
        """LocationsClient Instance

        :type: :class:`LocationsClient <hcloud.locations.client.LocationsClient>`
        """
        self.servers = ServersClient(self)
        """ServersClient Instance

        :type: :class:`ServersClient <hcloud.servers.client.ServersClient>`
        """
        self.server_types = ServerTypesClient(self)
        """ServerTypesClient Instance

        :type: :class:`ServerTypesClient <hcloud.server_types.client.ServerTypesClient>`
        """
        self.volumes = VolumesClient(self)
        """VolumesClient Instance

        :type: :class:`VolumesClient <hcloud.volumes.client.VolumesClient>`
        """
        self.actions = ActionsClient(self)
        """ActionsClient Instance

        :type: :class:`ActionsClient <hcloud.actions.client.ActionsClient>`
        """
        self.images = ImagesClient(self)
        """ImagesClient Instance

        :type: :class:`ImagesClient <hcloud.images.client.ImagesClient>`
        """
        self.isos = IsosClient(self)
        """ImagesClient Instance

        :type: :class:`IsosClient <hcloud.isos.client.IsosClient>`
        """
        self.ssh_keys = SSHKeysClient(self)
        """SSHKeysClient Instance

        :type: :class:`SSHKeysClient <hcloud.ssh_keys.client.SSHKeysClient>`
        """
        self.floating_ips = FloatingIPsClient(self)
        """FloatingIPsClient Instance

        :type: :class:`FloatingIPsClient <hcloud.floating_ips.client.FloatingIPsClient>`
        """
        self.primary_ips = PrimaryIPsClient(self)
        """PrimaryIPsClient Instance

        :type: :class:`PrimaryIPsClient <hcloud.primary_ips.client.PrimaryIPsClient>`
        """
        self.networks = NetworksClient(self)
        """NetworksClient Instance

        :type: :class:`NetworksClient <hcloud.networks.client.NetworksClient>`
        """
        self.certificates = CertificatesClient(self)
        """CertificatesClient Instance

        :type: :class:`CertificatesClient <hcloud.certificates.client.CertificatesClient>`
        """

        self.load_balancers = LoadBalancersClient(self)
        """LoadBalancersClient Instance

        :type: :class:`LoadBalancersClient <hcloud.load_balancers.client.LoadBalancersClient>`
        """

        self.load_balancer_types = LoadBalancerTypesClient(self)
        """LoadBalancerTypesClient Instance

        :type: :class:`LoadBalancerTypesClient <hcloud.load_balancer_types.client.LoadBalancerTypesClient>`
        """

        self.firewalls = FirewallsClient(self)
        """FirewallsClient Instance

        :type: :class:`FirewallsClient <hcloud.firewalls.client.FirewallsClient>`
        """

        self.placement_groups = PlacementGroupsClient(self)
        """PlacementGroupsClient Instance

        :type: :class:`PlacementGroupsClient <hcloud.placement_groups.client.PlacementGroupsClient>`
        """

    def _get_user_agent(self) -> str:
        """Get the user agent of the hcloud-python instance with the user application name (if specified)

        :return: The user agent of this hcloud-python instance
        """
        user_agents = []
        for name, version in [
            (self._application_name, self._application_version),
            (self.__user_agent_prefix, self._version),
        ]:
            if name is not None:
                user_agents.append(name if version is None else f"{name}/{version}")

        return " ".join(user_agents)

    def _get_headers(self) -> dict:
        headers = {
            "User-Agent": self._get_user_agent(),
            "Authorization": f"Bearer {self.token}",
        }
        return headers

    def _raise_exception_from_response(self, response: requests.Response) -> NoReturn:
        raise APIException(
            code=response.status_code,
            message=response.reason,
            details={"content": response.content},
        )

    def _raise_exception_from_content(self, content: dict) -> NoReturn:
        raise APIException(
            code=content["error"]["code"],
            message=content["error"]["message"],
            details=content["error"]["details"],
        )

    def request(  # type: ignore[no-untyped-def]
        self,
        method: str,
        url: str,
        tries: int = 1,
        **kwargs,
    ) -> dict:
        """Perform a request to the Hetzner Cloud API, wrapper around requests.request

        :param method: HTTP Method to perform the Request
        :param url: URL of the Endpoint
        :param tries: Tries of the request (used internally, should not be set by the user)
        :param timeout: Requests timeout in seconds
        :return: Response
        """
        timeout = kwargs.pop("timeout", self._requests_timeout)

        response = self._requests_session.request(
            method=method,
            url=self._api_endpoint + url,
            headers=self._get_headers(),
            timeout=timeout,
            **kwargs,
        )

        content = response.content
        try:
            if len(content) > 0:
                content = response.json()
        except (TypeError, ValueError):
            self._raise_exception_from_response(response)

        if not response.ok:
            if content:
                assert isinstance(content, dict)
                if content["error"]["code"] == "rate_limit_exceeded" and tries < 5:
                    time.sleep(tries * self._retry_wait_time)
                    tries = tries + 1
                    return self.request(method, url, tries, **kwargs)

                self._raise_exception_from_content(content)
            else:
                self._raise_exception_from_response(response)

        # TODO: return an empty dict instead of an empty string when content == "".
        return content  # type: ignore[return-value]
